# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import json
import uuid
import os
import sys
import logging
import warnings

import struct
import time

from google.protobuf import service as pb_service
from odps import compat
from odps.compat import six
from odps.errors import InvalidStateSetting, NoSuchObject

from . import errors
from .config import options
from .utils import get_property

try:
    from .proto import cupidtaskparam_pb2
except TypeError:
    warnings.warn('Cannot import protos from pycupid: '
        'consider upgrading your protobuf python package.', ImportWarning)
    raise ImportError

logger = logging.getLogger(__name__)
logger_mars = logging.getLogger('odps.mars_extension.cupid')

if sys.version_info[0] >= 3:
    b64decodebytes = base64.decodebytes
else:
    b64decodebytes = base64.decodestring


class CupidRpcController(pb_service.RpcController):
    def __init__(self):
        super(CupidRpcController, self).__init__()
        self._fail_msg = None

    def Reset(self):
        self._fail_msg = None

    def Failed(self):
        return self._fail_msg is not None

    def ErrorText(self):
        return self._fail_msg

    def StartCancel(self):
        pass

    def SetFailed(self, reason):
        self._fail_msg = reason

    def IsCanceled(self):
        return False

    def NotifyOnCancel(self, callback):
        pass


class CupidRpcChannel(pb_service.RpcChannel):
    @staticmethod
    def submit_job(param_pb, running_mode, priority=None, with_resource=False, session=None,
                   running_cluster=None, task_name='cupid_task'):
        from odps import ODPS, compat
        from odps.models import CupidTask

        if logger.getEffectiveLevel() <= logging.DEBUG:
            param_pb_str = str(param_pb)
            if isinstance(param_pb, cupidtaskparam_pb2.CupidTaskParam):
                for conf in param_pb.jobconf.jobconfitem:
                    if conf.key == 'odps.access.id':
                        param_pb_str = param_pb_str.replace(conf.value, '** access-id **')
                    elif conf.key == 'odps.access.key':
                        param_pb_str = param_pb_str.replace(conf.value, '** access-key **')
            logger.debug('Job param proto: %s', param_pb_str)

        odps = session.odps if session else ODPS.from_global()
        plan_string = param_pb.SerializeToString()
        res_kw = dict(fileobj=compat.BytesIO(plan_string))
        if not with_resource:
            res_kw['is_temp_resource'] = True
        res = odps.create_resource('cupid_plan_' + str(uuid.uuid4()), 'file', **res_kw)

        task_info = ','.join([res.name, odps.project, running_mode])

        props = dict()

        if options.cupid.application_type:
            props['odps.cupid.application.type'] = options.cupid.application_type
            props['odps.moye.runtime.type'] = options.cupid.application_type
        if options.biz_id:
            props['biz_id'] = options.biz_id
        if options.cupid.major_task_version:
            props['odps.task.major.version'] = options.cupid.major_task_version
        context_file = get_property('odps.exec.context.file')
        if context_file and os.path.exists(context_file):
            with open(context_file, 'r') as cf:
                file_settings = json.loads(cf.read()).get('settings', {})
            props.update(file_settings)

        task = CupidTask(task_name, task_info, props)
        inst = odps.get_project().instances.create(task=task, priority=priority, running_cluster=running_cluster)
        inst = odps.get_instance(inst.id)
        return inst

    @staticmethod
    def get_cupid_detail(inst):
        inst.reload()

        params = {'instancedetail': '', 'taskname': 'cupid_task'}
        resp = inst._client.get(inst.resource(), params=params)
        if resp.content == b'{}':
            logger.debug('Empty content retrieved. Use get_task_detail2 first.')
            inst.get_task_detail2('cupid_task')
            resp = inst._client.get(inst.resource(), params=params)

        return resp.content

    @classmethod
    def get_cupid_status(cls, inst):
        details = cls.get_cupid_detail(inst)
        if details.startswith(b'Failed') and b'recycled' in details:
            raise errors.InstanceRecycledError(details)
        details_pb = cupidtaskparam_pb2.CupidTaskDetailResultParam()
        details_pb.ParseFromString(b64decodebytes(details))

        status_keys = ('ready', 'waiting', 'running', 'success', 'failed', 'cancelled')
        if not any(details_pb.HasField(sk) for sk in status_keys):
            details_pb.ready.CopyFrom(cupidtaskparam_pb2.Ready())
        return details_pb

    @classmethod
    def wait_cupid_instance(cls, inst, state=None, master_timeout=120):
        state = state or 'success'
        wait_keys = ('running', 'success', 'ready')
        sleep_time = 0

        while True:
            result = cls.get_cupid_status(inst)
            sleep_time = min(5, sleep_time + 1)
            if result.HasField('failed'):
                if result.failed.HasField('cupidTaskFailed'):
                    msg = result.failed.cupidTaskFailed.cupidTaskFailedMsg
                else:
                    msg = result.failed.bizFailed.bizFailedMsg
                if msg.startswith("runTask failed:") or msg.startswith("app run failed!"):
                    raise errors.CupidUserError(msg)
                else:
                    raise errors.CupidError(msg)
            elif result.HasField('cancelled'):
                raise errors.CupidUserError('Instance canceled.')
            elif result.HasField(state):
                return getattr(getattr(result, state), state + 'Msg')
            elif any(result.HasField(wk) for wk in wait_keys):
                time.sleep(sleep_time)
            else:
                msg_pb = str(result)
                if msg_pb != '':
                    logger.warning('Unexpected status: %s' % msg_pb)
                time.sleep(sleep_time)


class CupidTaskServiceRpcChannel(CupidRpcChannel):
    def __init__(self, session):
        super(CupidTaskServiceRpcChannel, self).__init__()
        self.cupid_session = session

    def CallMethod(self, method, rpc_controller,
                   request, response_class, done):
        task_service_req = cupidtaskparam_pb2.TaskServiceRequest(
            methodName=method.name,
            requestInBytes=request.SerializeToString(),
        )
        job_conf = self.cupid_session.job_conf()
        task_operator = cupidtaskparam_pb2.CupidTaskOperator(moperator='TaskServiceRequest')
        task_param = cupidtaskparam_pb2.CupidTaskParam(
            mcupidtaskoperator=task_operator, jobconf=job_conf,
            taskServiceRequest=task_service_req
        )
        inst = self.submit_job(task_param, 'eAsyncNotFuxiJob', session=self.cupid_session)
        logger_mars.debug('Cupid task instance: %s, method: %s' % (inst.id, method.name))

        resp_str = self.wait_cupid_instance(inst)
        if isinstance(resp_str, six.text_type):
            resp_str = resp_str.encode()
        resp = response_class()
        resp.ParseFromString(b64decodebytes(resp_str))
        return resp


class SandboxRpcChannel(pb_service.RpcChannel):
    def CallMethod(self, method, controller, request, response_class, done):
        from .runtime import context

        context = context()

        try:
            sio = compat.BytesIO()
            sio.write(struct.pack('<I', method.index))
            sio.write(request.SerializeToString())

            logger.debug('SandboxRpcChannel sync_call service: %s, method id: %d request: %s',
                         method.containing_service.full_name, method.index, request)
            res = context.channel_client.sync_call(method.containing_service.full_name,
                                                   sio.getvalue())
            resp = response_class()
            resp.ParseFromString(res)
            logger.debug('SandboxRpcChannel sync_call result: %s', resp)
            return resp
        except Exception as exc:
            logger.exception('SandboxRpcChannel CallMethod fail: %s', str(exc))
            controller.SetFailed(str(exc))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..rewriter import BaseRewriter
from ...expr.reduction import Cat, GroupedCat
from ...expr.window import *
from ...expr.merge import *
from ...expr.expressions import LateralViewCollectionExpr
from ...expr.element import Func
from ...expr.utils import get_attrs
from ....errors import NoSuchObject
from ...utils import is_source_collection
from ....models import Table


class Rewriter(BaseRewriter):
    def visit_project_collection(self, expr):
        self._rewrite_reduction_in_projection(expr)

    def visit_filter_collection(self, expr):
        self._rewrite_reduction_in_filter(expr)

    def visit_join(self, expr):
        if expr._predicate and isinstance(expr._predicate, list):
            expr._predicate = reduce(operator.and_, expr._predicate)

        for node in (expr.rhs,):
            parents = self._parents(node)
            if isinstance(node, JoinCollectionExpr):
                projection = JoinProjectCollectionExpr(
                    _input=node, _schema=node.schema,
                    _fields=node._fetch_fields())
                self._sub(node, projection, parents)
            elif isinstance(node, JoinProjectCollectionExpr):
                self._sub(node.input, node, parents)

        need_project = [False, ]

        def walk(node):
            if isinstance(node, JoinCollectionExpr) and \
                    node.column_conflict:
                need_project[0] = True
                return

            if isinstance(node, JoinCollectionExpr):
                walk(node.lhs)

        walk(expr)

        if need_project[0]:
            parents = self._parents(expr)
            if not parents or \
                    not any(isinstance(parent, (ProjectCollectionExpr, JoinCollectionExpr))
                            for parent in parents):
                to_sub = expr[expr]
                self._sub(expr, to_sub, parents)

    def visit_lateral_view(self, expr):
        parents = self._parents(expr)
        if not parents \
                or not any(isinstance(parent, ProjectCollectionExpr) \
                           and not isinstance(parent, LateralViewCollectionExpr) for parent in parents):
            to_sub = ProjectCollectionExpr(
                _input=expr, _schema=expr.schema, _fields=expr._fields
            )
            self._sub(expr, to_sub, parents)

    def _handle_function(self, expr, raw_inputs):
        # Since Python UDF cannot support decimal field,
        # We will try to replace the decimal input with string.
        # If the output is decimal, we will also try to replace it with string,
        # and then cast back to decimal
        def no_output_decimal():
            if isinstance(expr, (SequenceExpr, Scalar)):
                return expr.dtype != types.decimal
            else:
                return all(t != types.decimal for t in expr.schema.types)

        if isinstance(expr, Func):
            return

        if all(input.dtype != types.decimal for input in raw_inputs) and \
                no_output_decimal():
            return

        inputs = list(raw_inputs)
        for input in raw_inputs:
            if input.dtype == types.decimal:
                self._sub(input, input.astype('string'), parents=[expr, ])
        if hasattr(expr, '_raw_inputs'):
            expr._raw_inputs = inputs
        else:
            assert len(inputs) == 1
            expr._raw_input = inputs[0]

        attrs = get_attrs(expr)
        attr_values = dict((attr, getattr(expr, attr, None)) for attr in attrs)
        if isinstance(expr, (SequenceExpr, Scalar)):
            if expr.dtype == types.decimal:
                if isinstance(expr, SequenceExpr):
                    attr_values['_data_type'] = types.string
                    attr_values['_source_data_type'] = types.string
                else:
                    attr_values['_value_type'] = types.string
                    attr_values['_source_value_type'] = types.string
            sub = type(expr)._new(**attr_values)

            if expr.dtype == types.decimal:
                sub = sub.astype('decimal')
        else:
            names = expr.schema.names
            tps = expr.schema.types
            cast_names = set()
            if any(tp == types.decimal for tp in tps):
                new_tps = []
                for name, tp in zip(names, tps):
                    if tp != types.decimal:
                        new_tps.append(tp)
                        continue
                    new_tps.append(types.string)
                    cast_names.add(name)
                if len(cast_names) > 0:
                    attr_values['_schema'] = TableSchema.from_lists(names, new_tps)
            sub = type(expr)(**attr_values)

            if len(cast_names) > 0:
                fields = []
                for name in names:
                    if name in cast_names:
                        fields.append(sub[name].astype('decimal'))
                    else:
                        fields.append(name)
                sub = sub[fields]

        self._sub(expr, sub)

    def visit_function(self, expr):
        self._handle_function(expr, expr._inputs)

    def visit_reshuffle(self, expr):
        if isinstance(expr.input, JoinCollectionExpr):
            sub = JoinProjectCollectionExpr(
                _input=expr.input, _schema=expr.input.schema,
                _fields=expr.input._fetch_fields())
            self._sub(expr.input, sub)

    def visit_apply_collection(self, expr):
        if (
            isinstance(expr._input, JoinCollectionExpr)
            and (expr._input._mapjoin or expr._input._skewjoin)
        ):
            node = expr._input
            projection = JoinProjectCollectionExpr(
                _input=node, _schema=node.schema,
                _fields=node._fetch_fields())
            self._sub(node, projection)
        self._handle_function(expr, expr._fields)

    def visit_user_defined_aggregator(self, expr):
        self._handle_function(expr, [expr.input, ])

    def visit_column(self, expr):
        if is_source_collection(expr.input) and isinstance(expr._input._source_data, Table):
            try:
                if expr.input._source_data.table_schema.is_partition(expr.source_name) and \
                                expr.dtype != types.string:
                    expr._source_data_type = types.string
            except NoSuchObject:
                return

    def visit_reduction(self, expr):
        if isinstance(expr, (Cat, GroupedCat)):
            if expr._na_rep is not None:
                input = expr.input.fillna(expr._na_rep)
                self._sub(expr.input, input, parents=(expr, ))


# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .utils import metrics_result


def _run_evaluation_node(df, model, cols, execute_now=True, result_callback=None):
    from . import _customize
    eval_fun = getattr(_customize, '_eval_clustering')
    return eval_fun(df, model, cols=cols, execute_now=execute_now, _result_callback=result_callback)


@metrics_result(_run_evaluation_node)
def calinhara_score(df, model, cols=None):
    r"""
    Calculate Calinski-Harabasz coefficient for a clustering model.

    Calinski-Harabasz coefficient, also known as VRC (variance ratio criterion), is defined as

    .. math::

        VRC_k=\frac{SS_B}{SS_W}\times\frac{n-k}{k-1}

    Where :math:`SS_B` is the inter-cluster variance matrix, :math:`SS_W` is the intra-cluster variance matrix, while :math:`N` is the number of instances and :math:`k` is the number of cluster centers.

    :math:`SS_B` is defined as follows

    .. math::

        SS_B=\sum_{i=1}^k n_i \left\lVert m_i - \overline{m} \right\rVert^2

    While :math:`SS_W` is defined as follows

    .. math::

        SS_W = \sum_{i=1}^k \sum_{x \in c_i} \left\lVert x - m_i \right\rVert^2

    :param df: input DataFrame
    :param model: model generated by a clustering algorithm. Currently only k-Means is supported.
    :param cols: selected columns. Feature columns on ``df`` by default.
    """
    return _run_evaluation_node(df, model, cols)['calinhara']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..sqlalchemy.engine import SQLAlchemyEngine
from ..odpssql.types import df_schema_to_odps_schema
from .compiler import SeahawksCompiler
from .models import SeahawksTable


class SeahawksEngine(SQLAlchemyEngine):

    def __init__(self, odps):
        super(SeahawksEngine, self).__init__(odps=odps)

    def _compile(self, expr, dag):
        compiler = SeahawksCompiler(dag, self._odps)
        return compiler.compile(expr)

    def _status_ui(self, ui):
        ui.status('Try to execute by seahawks...')

    @classmethod
    def _create_table(cls, table_name, sa, expr):
        from .ext import CreateTempTableAs

        return CreateTempTableAs(table_name, sa, expr)

    def _get_table(self, table_name, expr_dag, bind=None):
        return SeahawksTable(name=table_name,
                             schema=expr_dag.root.schema)



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import logging.config
import itertools
import platform
import warnings
try:
    if sys.version_info[:2] < (3, 3):
        import xml.etree.cElementTree as ElementTree
    else:
        import xml.etree.ElementTree as ElementTree
except ImportError:
    import xml.etree.ElementTree as ElementTree
try:
    ElementTreeParseError = getattr(ElementTree, 'ParseError')
except AttributeError:
    ElementTreeParseError = getattr(ElementTree, 'XMLParserError')
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable
from unicodedata import east_asian_width

from .lib import six

PY27 = six.PY2 and sys.version_info[1] == 7
LESS_PY32 = sys.version_info[:2] < (3, 2)
LESS_PY33 = sys.version_info[:2] < (3, 3)
LESS_PY34 = sys.version_info[:2] < (3, 4)
LESS_PY35 = sys.version_info[:2] < (3, 5)
PYPY = platform.python_implementation().lower() == 'pypy'

SEEK_SET = 0
SEEK_CUR = 1
SEEK_END = 2

# Definition of East Asian Width
# http://unicode.org/reports/tr11/
# Ambiguous width can be changed by option
_EAW_MAP = {'Na': 1, 'N': 1, 'W': 2, 'F': 2, 'H': 1}

import decimal
DECIMAL_TYPES = [decimal.Decimal, ]

import json  # don't remove

try:
    TimeoutError = TimeoutError
except NameError:
    TimeoutError = type('TimeoutError', (RuntimeError,), {})


if six.PY3:
    lrange = lambda *x: list(range(*x))
    lzip = lambda *x: list(zip(*x))
    lkeys = lambda x: list(x.keys())
    lvalues = lambda x: list(x.values())
    litems = lambda x: list(x.items())

    irange = range
    izip = zip

    long_type = int

    import io
    StringIO = io.StringIO
    BytesIO = io.BytesIO

    if LESS_PY34:
        from .lib import enum
    else:
        import enum

    if LESS_PY33:
        try:
            import cdecimal as decimal

            DECIMAL_TYPES.append(decimal.Decimal)
        except:
            import decimal
    else:
        import decimal

    def u(s):
        return s

    def strlen(data, encoding=None):
        # encoding is for compat with PY2
        return len(data)

    def east_asian_len(data, encoding=None, ambiguous_width=1):
        """
        Calculate display width considering unicode East Asian Width
        """
        if isinstance(data, six.text_type):
            return sum([_EAW_MAP.get(east_asian_width(c), ambiguous_width) for c in data])
        else:
            return len(data)

    dictconfig = lambda config: logging.config.dictConfig(config)

    import builtins
    from concurrent import futures  # don't remove

    from .lib import cgi_compat as cgi

    UnsupportedOperation = io.UnsupportedOperation

    try:
        from .lib.version import Version
    except BaseException:
        from distutils.version import LooseVersion as Version

    from threading import Semaphore
else:
    lrange = range
    lzip = zip
    lkeys = lambda x: x.keys()
    lvalues = lambda x: x.values()
    litems = lambda x: x.items()

    irange = xrange
    izip = itertools.izip

    long_type = long

    from .lib import enum

    try:
        import cdecimal as decimal
        DECIMAL_TYPES.append(decimal.Decimal)
    except ImportError:
        import decimal

    try:
        import cStringIO as StringIO
    except ImportError:
        import StringIO
    StringIO = BytesIO = StringIO.StringIO

    def u(s):
        return unicode(s, "unicode_escape")

    def strlen(data, encoding=None):
        try:
            data = data.decode(encoding)
        except UnicodeError:
            pass
        return len(data)

    def east_asian_len(data, encoding=None, ambiguous_width=1):
        """
        Calculate display width considering unicode East Asian Width
        """
        if isinstance(data, six.text_type):
            try:
                data = data.decode(encoding)
            except UnicodeError:
                pass
            return sum([_EAW_MAP.get(east_asian_width(c), ambiguous_width) for c in data])
        else:
            return len(data)

    dictconfig = lambda config: logging.config.dictConfig(config)

    import __builtin__ as builtins  # don't remove
    from .lib import futures  # don't remove
    import cgi

    UnsupportedOperation = type("UnsupportedOperation", (OSError, ValueError), {})

    from distutils.version import LooseVersion as Version

    from threading import _Semaphore as _PySemaphore
    from .lib.monotonic import monotonic


    class Semaphore(_PySemaphore):
        def acquire(self, blocking=True, timeout=None):
            if not blocking and timeout is not None:
                raise ValueError("can't specify timeout for non-blocking acquire")
            rc = False
            endtime = None
            with self.__cond:
                while self.__value == 0:
                    if not blocking:
                        break
                    if timeout is not None:
                        if endtime is None:
                            endtime = monotonic() + timeout
                        else:
                            timeout = endtime - monotonic()
                            if timeout <= 0:
                                break
                    self.__cond.wait(timeout)
                else:
                    self.__value -= 1
                    rc = True
            return rc

if LESS_PY32:
    try:
        from .tests.dictconfig import dictConfig
        dictconfig = lambda config: dictConfig(config)
    except ImportError:
        pass

if six.PY3:
    from contextlib import suppress
else:
    from contextlib import contextmanager

    @contextmanager
    def suppress(*exceptions):
        try:
            yield
        except exceptions:
            pass

Enum = enum.Enum
DECIMAL_TYPES = tuple(DECIMAL_TYPES)
Decimal = decimal.Decimal

try:
    import pandas as pd
    if not hasattr(pd.DataFrame, 'sort_values'):
        pd.DataFrame.sort_values = pd.DataFrame.sort

    from pandas.core.internals import blocks as pd_blocks
    if not hasattr(pd_blocks, "new_block"):
        pd_blocks.new_block = pd_blocks.make_block

    if not hasattr(pd.RangeIndex, 'start'):
        pd.RangeIndex.start = property(fget=lambda x: x._start)
        pd.RangeIndex.stop = property(fget=lambda x: x._stop)
        pd.RangeIndex.step = property(fget=lambda x: x._step)
except ImportError:
    pass

from .lib.lib_utils import isvalidattr, dir2, getargspec, getfullargspec

from .lib.six.moves import reduce
from .lib.six.moves import reload_module
from .lib.six.moves.queue import Queue, Empty
from .lib.six.moves.urllib.request import urlretrieve
from .lib.six.moves import cPickle as pickle
from .lib.six.moves.urllib.parse import urlencode, urlparse, unquote, quote, quote_plus, parse_qsl
from .lib.six.moves import configparser as ConfigParser

from .lib.ext_types import Monthdelta


import datetime


class _FixedOffset(datetime.tzinfo):
    """
    A class building tzinfo objects for fixed-offset time zones.
    Note that FixedOffset(0, "UTC") is a different way to build a
    UTC tzinfo object.
    """
    def __init__(self, offset, name=None):
        self.__offset = datetime.timedelta(minutes=offset)
        self.__name = name

    def utcoffset(self, dt):
        return self.__offset

    def tzname(self, dt):
        return self.__name

    def dst(self, dt):
        return _ZERO_TIMEDELTA


try:
    import zoneinfo
    utc = zoneinfo.ZoneInfo("UTC")
    FixedOffset = _FixedOffset
except ImportError:
    try:
        import pytz
        utc = pytz.utc
        FixedOffset = pytz._FixedOffset
    except ImportError:
        _ZERO_TIMEDELTA = datetime.timedelta(0)
        FixedOffset = _FixedOffset
        utc = FixedOffset(0, 'UTC')


try:
    from email.utils import parsedate_to_datetime
except ImportError:
    import datetime
    from email.utils import parsedate_tz

    def parsedate_to_datetime(data):
        dt_tuple_with_tz = parsedate_tz(data)
        dtuple = dt_tuple_with_tz[:-1]
        tz = dt_tuple_with_tz[-1]
        if tz is None:
            return datetime.datetime(*dtuple[:6])
        return datetime.datetime(*dtuple[:6], tzinfo=FixedOffset(tz / 60.0))


__all__ = ['sys', 'builtins', 'logging.config', 'dictconfig', 'suppress',
           'reduce', 'reload_module', 'Queue', 'Empty', 'ElementTree', 'ElementTreeParseError',
           'urlretrieve', 'pickle', 'urlencode', 'urlparse', 'unquote', 'quote', 'quote_plus', 'parse_qsl',
           'Enum', 'ConfigParser', 'decimal', 'Decimal', 'DECIMAL_TYPES', 'FixedOffset', 'utc', 'Monthdelta',
           'Iterable', 'TimeoutError', 'cgi', 'parsedate_to_datetime', 'Version', 'Semaphore']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
from collections import OrderedDict

from ....models import TableSchema
from .... import types as odps_types
from ... import types as df_types
from ....compat import six
from ....config import options


_odps_to_df_types = {
    odps_types.tinyint: df_types.int8,
    odps_types.smallint: df_types.int16,
    odps_types.int_: df_types.int32,
    odps_types.bigint: df_types.int64,
    odps_types.float_: df_types.float32,
    odps_types.double: df_types.float64,
    odps_types.string: df_types.string,
    odps_types.datetime: df_types.datetime,
    odps_types.boolean: df_types.boolean,
    odps_types.binary: df_types.binary,
    odps_types.timestamp: df_types.timestamp,
    odps_types.timestamp_ntz: df_types.timestamp,
    odps_types.date: df_types.date,
    odps_types.json: df_types.json,
}

_df_to_odps_types = {
    df_types.int8: odps_types.bigint,
    df_types.int16: odps_types.bigint,
    df_types.int32: odps_types.bigint,
    df_types.int64: odps_types.bigint,
    df_types.float32: odps_types.double,
    df_types.float64: odps_types.double,
    df_types.boolean: odps_types.boolean,
    df_types.string: odps_types.string,
    df_types.datetime: odps_types.datetime,
    df_types.binary: odps_types.string,
    df_types.timestamp: odps_types.timestamp,
    df_types.date: odps_types.date,
    df_types.json: odps_types.json,
}

_df_to_odps_types2 = {
    df_types.int8: odps_types.tinyint,
    df_types.int16: odps_types.smallint,
    df_types.int32: odps_types.int_,
    df_types.int64: odps_types.bigint,
    df_types.float32: odps_types.float_,
    df_types.float64: odps_types.double,
    df_types.boolean: odps_types.boolean,
    df_types.string: odps_types.string,
    df_types.datetime: odps_types.datetime,
    df_types.binary: odps_types.binary,
    df_types.timestamp: odps_types.timestamp,
    df_types.date: odps_types.date,
    df_types.json: odps_types.json,
}


def is_odps_type(typ):
    return typ in _odps_to_df_types


def odps_type_to_df_type(odps_type):
    if isinstance(odps_type, six.string_types):
        odps_type = odps_types.validate_data_type(odps_type)

    if odps_type in _odps_to_df_types:
        return _odps_to_df_types[odps_type]
    elif isinstance(odps_type, odps_types.Decimal):
        return df_types.decimal
    elif isinstance(odps_type, (odps_types.Varchar, odps_types.Char)):
        return df_types.string
    elif isinstance(odps_type, odps_types.Array):
        return df_types.List(odps_type_to_df_type(odps_type.value_type))
    elif isinstance(odps_type, odps_types.Map):
        return df_types.Dict(odps_type_to_df_type(odps_type.key_type),
                             odps_type_to_df_type(odps_type.value_type))
    elif isinstance(odps_type, odps_types.Struct):
        type_dict = OrderedDict(
            (k, odps_type_to_df_type(v)) for k, v in odps_type.field_types.items()
        )
        return df_types.Struct(type_dict)
    else:
        raise KeyError(repr(odps_type))


def odps_schema_to_df_schema(odps_schema):
    names = [col.name for col in odps_schema.columns]
    types = [odps_type_to_df_type(col.type) for col in odps_schema.columns]

    return TableSchema.from_lists(names, types)


def df_type_to_odps_type(df_type, use_odps2_types=None, project=None):
    if use_odps2_types is None:
        use_odps2_types = get_local_use_odps2_types(project)

    if use_odps2_types:
        df_to_odps_types = _df_to_odps_types2
    else:
        df_to_odps_types = _df_to_odps_types
    if isinstance(df_type, six.string_types):
        df_type = df_types.validate_data_type(df_type)

    if df_type in df_to_odps_types:
        return df_to_odps_types[df_type]
    elif df_type == df_types.decimal:
        return odps_types.Decimal()
    elif isinstance(df_type, df_types.List):
        return odps_types.Array(
            df_type_to_odps_type(df_type.value_type, use_odps2_types=use_odps2_types)
        )
    elif isinstance(df_type, df_types.Dict):
        return odps_types.Map(
            df_type_to_odps_type(df_type.key_type, use_odps2_types=use_odps2_types),
            df_type_to_odps_type(df_type.value_type, use_odps2_types=use_odps2_types),
        )
    elif isinstance(df_type, df_types.Struct):
        type_dict = OrderedDict(
            (k, df_type_to_odps_type(v, use_odps2_types=use_odps2_types))
            for k, v in df_type.field_types.items()
        )
        return odps_types.Struct(type_dict)
    else:
        raise KeyError(repr(df_type))


def df_schema_to_odps_schema(df_schema, ignorecase=False, project=None):
    names = [col.name.lower() if ignorecase else col.name
             for col in df_schema._columns]
    types = [
        df_type_to_odps_type(col.type, project=project)
        for col in df_schema._columns
    ]

    partition_names, partition_types = None, None
    if df_schema.partitions:
        partition_names = [
            col.name.lower() if ignorecase else col.name
            for col in df_schema.partitions
        ]
        partition_types = [
            df_type_to_odps_type(col.type, project=project)
            for col in df_schema.partitions
        ]

    return TableSchema.from_lists(
        names, types, partition_names=partition_names, partition_types=partition_types
    )



_use_odps2_types_local = threading.local()


def set_local_use_odps2_types(val):
    _use_odps2_types_local.value = val


def get_local_use_odps2_types(project=None):
    if options.sql.use_odps2_extension is not None:
        ret = bool(options.sql.use_odps2_extension)
    elif project is not None:
        project_prop = project.properties.get("odps.sql.type.system.odps2")
        ret = ("true" == (project_prop or "false").lower())
        set_local_use_odps2_types(ret)
    elif getattr(_use_odps2_types_local, "value", None) is not None:
        ret = _use_odps2_types_local.value
    else:
        ret = False
    return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json  # don't remove
import random
import sys
import time
import warnings
from collections import OrderedDict

from .core import AbstractXMLRemoteModel
from .. import serializers, errors, utils
from ..compat import enum, six
from ..config import options


class Task(AbstractXMLRemoteModel):

    __slots__ = 'name', 'comment', 'properties'

    _type_indicator = 'type'

    name = serializers.XMLNodeField('Name')
    type = serializers.XMLTagField('.')
    comment = serializers.XMLNodeField('Comment')
    properties = serializers.XMLNodePropertiesField('Config', 'Property',
                                                    key_tag='Name', value_tag='Value')

    def __new__(cls, *args, **kwargs):
        typo = kwargs.get('type')

        if typo is not None:
            task_cls = None
            for v in six.itervalues(globals()):
                if not isinstance(v, type) or not issubclass(v, Task):
                    continue
                cls_type = getattr(v, '_root', v.__name__)
                if typo == cls_type:
                    task_cls = v
            if task_cls is None:
                task_cls = cls
        else:
            task_cls = cls

        return object.__new__(task_cls)

    def set_property(self, key, value):
        if self.properties is None:
            self.properties = OrderedDict()
        self.properties[key] = value

    def _update_property_json(self, field, value):
        def update(kv, dest):
            if not kv:
                return
            for k, v in six.iteritems(kv):
                if isinstance(v, bool):
                    dest[k] = 'true' if v else 'false'
                else:
                    dest[k] = str(v)

        if self.properties is None:
            self.properties = OrderedDict()
        if field in self.properties:
            settings = json.loads(self.properties[field])
        else:
            settings = OrderedDict()
        update(value, settings)
        self.properties[field] = json.dumps(settings)

    def update_settings(self, value):
        self._update_property_json('settings', value)

    def serialize(self):
        if type(self) is Task:
            raise errors.ODPSError('Unknown task type')
        return super(Task, self).serialize()

    @property
    def instance(self):
        return self.parent.parent

    @property
    def progress(self):
        """
        Get progress of a task.
        """
        return self.instance.get_task_progress(self.name)

    @property
    def stages(self):
        """
        Get execution stages of a task.
        """
        return self.instance.get_task_progress(self.name).stages

    @property
    def result(self):
        """
        Get execution result of the task.
        """
        return self.instance.get_task_result(self.name)

    @property
    def summary(self):
        """
        Get execution summary of the task.
        """
        return self.instance.get_task_summary(self.name)

    @property
    def detail(self):
        """
        Get execution details of the task.
        """
        return self.instance.get_task_detail(self.name)

    @property
    def quota(self):
        """
        Get quota json of the task.
        """
        return self.instance.get_task_quota(self.name)

    @property
    def workers(self):
        """
        Get workers of the task.
        """
        return self.instance.get_task_workers(self.name)

    def get_info(self, key):
        """
        Get associated information of the task.
        """
        return self.instance.get_task_info(self.name, key)

    def put_info(self, key, value):
        """
        Put associated information of the task.
        """
        return self.instance.put_task_info(self.name, key, value)


def format_cdata(query, semicolon=False):
    stripped_query = query.strip()
    if semicolon and not stripped_query.endswith(';'):
        stripped_query += ';'
    return '<![CDATA[%s]]>' % stripped_query


def collect_sql_settings(value, glob):
    from .. import __version__

    settings = OrderedDict()
    if options.default_task_settings:
        settings = options.default_task_settings

    settings["PYODPS_VERSION"] = __version__
    settings["PYODPS_PYTHON_VERSION"] = sys.version

    if glob:
        if options.sql.use_odps2_extension:
            settings['odps.sql.type.system.odps2'] = True
        if options.local_timezone is not None:
            if not options.local_timezone:
                settings['odps.sql.timezone'] = 'Etc/GMT'
            elif isinstance(options.local_timezone, bool):
                from ..lib import tzlocal

                zone = tzlocal.get_localzone()
                settings['odps.sql.timezone'] = utils.get_zone_name(zone)
            elif isinstance(options.local_timezone, six.string_types):
                settings['odps.sql.timezone'] = options.local_timezone
            else:
                zone = options.local_timezone
                zone_str = utils.get_zone_name(zone)
                if zone_str is None:
                    warnings.warn('Failed to get timezone string from options.local_timezone. '
                                  'You need to deal with timezone in the return data yourself.')
                else:
                    settings['odps.sql.timezone'] = zone_str
        if options.sql.settings:
            settings.update(options.sql.settings)
    if value:
        settings.update(value)
    return settings


class SQLTask(Task):
    __slots__ = '_anonymous_sql_task_name',

    _root = 'SQL'
    _anonymous_sql_task_name = 'AnonymousSQLTask'

    query = serializers.XMLNodeField('Query',
                                     serialize_callback=lambda s: format_cdata(s, True))

    def __init__(self, **kwargs):
        if 'name' not in kwargs:
            kwargs['name'] = SQLTask._anonymous_sql_task_name
        super(SQLTask, self).__init__(**kwargs)

    def serial(self):
        if self.properties is None:
            self.properties = OrderedDict()

        key = 'settings'
        if key not in self.properties:
            self.properties[key] = '{"odps.sql.udf.strict.mode": "true"}'

        return super(SQLTask, self).serial()

    def update_sql_settings(self, value=None, glob=True):
        settings = collect_sql_settings(value, glob)
        self.update_settings(settings)

    def update_aliases(self, value):
        self._update_property_json('aliases', value)

    @property
    def warnings(self):
        return json.loads(self.get_info('warnings')).get('warnings')


class MergeTask(Task):
    _root = 'Merge'

    table = serializers.XMLNodeField('TableName')

    def __init__(self, name=None, **kwargs):
        if name is None:
            name = 'merge_task_{0}_{1}'.format(int(time.time()), random.randint(100000, 999999))
        kwargs['name'] = name
        super(MergeTask, self).__init__(**kwargs)


class CupidTask(Task):
    _root = 'CUPID'

    plan = serializers.XMLNodeField('Plan', serialize_callback=format_cdata)

    def __init__(self, name=None, plan=None, hints=None, **kwargs):
        kwargs['name'] = name
        kwargs['plan'] = plan
        super(CupidTask, self).__init__(**kwargs)
        hints = hints or {}
        self.set_property('type', 'cupid')
        if hints:
            self.set_property('settings', json.dumps(hints))


class SQLCostTask(Task):
    __slots__ = '_anonymous_sql_cost_task_name',

    _root = 'SQLCost'
    _anonymous_sql_cost_task_name = 'AnonymousSQLCostTask'

    query = serializers.XMLNodeField('Query',
                                     serialize_callback=lambda s: format_cdata(s, True))

    def __init__(self, **kwargs):
        if 'name' not in kwargs:
            kwargs['name'] = self._anonymous_sql_cost_task_name
        super(SQLCostTask, self).__init__(**kwargs)

    def update_sql_cost_settings(self, value=None, glob=True):
        settings = collect_sql_settings(value, glob)
        self.update_settings(settings)


class SQLRTTask(Task):
    _root = "SQLRT"

    def update_sql_rt_settings(self, value=None, glob=True):
        settings = collect_sql_settings(value, glob)
        self.update_settings(settings)


class MaxFrameTask(Task):
    __slots__ = ("_output_format", "_major_version", "_service_endpoint")
    _root = "MaxFrame"
    _anonymous_task_name = "AnonymousMaxFrameTask"

    class CommandType(enum.Enum):
        CREATE_SESSION = "CREATE_SESSION"
        PYTHON_PACK = "PYTHON_PACK"

    command = serializers.XMLNodeField(
        "Command",
        default=CommandType.CREATE_SESSION,
        parse_callback=lambda t: MaxFrameTask.CommandType(t.upper()),
        serialize_callback=lambda t: t.value,
    )

    def __init__(self, **kwargs):
        kwargs["name"] = kwargs.get("name") or self._anonymous_task_name
        self._major_version = kwargs.pop("major_version", None)
        self._service_endpoint = kwargs.pop("service_endpoint", None)
        super(MaxFrameTask, self).__init__(**kwargs)

        if self.properties is None:
            self.properties = OrderedDict()
        self.properties["settings"] = "{}"

    def serial(self):
        if options.default_task_settings:
            settings = options.default_task_settings.copy()
        else:
            settings = OrderedDict()

        if self._major_version is not None:
            settings["odps.task.major.version"] = self._major_version
        if self._service_endpoint is not None:
            settings["odps.service.endpoint"] = self._service_endpoint

        if "settings" in self.properties:
            settings.update(json.loads(self.properties["settings"]))

        self.properties["settings"] = json.dumps(settings)
        return super(MaxFrameTask, self).serial()


try:
    from ..internal.models.tasks import *  # noqa: F401
except ImportError:
    pass



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libc.stdint cimport *
from libc.string cimport *
from libcpp.vector cimport vector

from ...src.types_c cimport SchemaSnapshot
from ..checksum_c cimport Checksum
from .util_c cimport *
from .block_encoder_c cimport Encoder


ctypedef void (*_NOGIL_WRITER)(BasePandasWriter self, ArrayVariantPtrs &record, int row) nogil


cdef class BasePandasWriter:
    cdef object _schema
    cdef object _columns
    cdef object _stream
    cdef Checksum _crc
    cdef Checksum _crccrc
    cdef int _n_columns
    cdef int _read_limit
    cdef Encoder _encoder

    cdef object _mem_cache
    cdef object _mem_cache_view
    cdef int _mem_cache_size

    cdef int _count
    cdef uint32_t _row_pos

    cdef vector[_NOGIL_WRITER] _nogil_writers

    cdef void _write_long_val(self, long val) nogil

    cdef void _write_long(self, ArrayVariantPtrs &aptr, int index) nogil

    cdef void _write_bool(self, ArrayVariantPtrs &aptr, int index) nogil

    cdef void _write_float(self, ArrayVariantPtrs &aptr, int index) nogil

    cdef void _write_double(self, ArrayVariantPtrs &aptr, int index) nogil

    cdef int _write_single_ndarray_nogil(
            self, ArrayVariantPtrs &col_ptr, vector[int] &dims, vector[int] &col_to_dim,
            long start_pos, long limit, vector[long] &dim_offsets) nogil
    cdef int _write_dims_nogil(
            self, vector[ArrayVariantPtrs] &col_ptrs, vector[int] &col_to_dim,
            long start_row, long limit) nogil

    cpdef init_cache(self)
    cpdef reset_positions(self)
    cpdef write_stream(self, object data, int length)

    cpdef _write_single_array(self, object data, object columns, long limit, object dim_offsets)
    cpdef _write_dims(self, object data, object columns, long limit)
    cpdef write(self, object data, object columns=*, long limit=*, object dim_offsets=*)

    cpdef flush(self)
    cpdef close(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from io import BytesIO

import numpy as np
from mars import opcodes as OperandDef
from mars.serialize import (
    ValueType,
    ListField,
    TupleField,
    StringField,
    KeyField,
    BoolField,
)
from mars.tensor.array_utils import as_np_array
from mars.tensor.datastore.core import TensorDataStore
from mars.lib.sparse import SparseNDArray


class TensorStoreCOO(TensorDataStore):
    _op_type_ = OperandDef.STORE_COO

    _input = KeyField("input")
    _path = StringField("path")
    _dim_cols = ListField("dim_cols", ValueType.string)
    _value_col = StringField("value_col")
    _storage_options = StringField("storage_options")
    _global_index = BoolField("global_index", default=False)
    _axis_offsets = TupleField("axis_offsets")

    def __init__(
        self,
        dtype=None,
        path=None,
        dim_cols=None,
        value_col=None,
        storage_options=None,
        sparse=True,
        global_index=False,
        **kw
    ):
        super(TensorStoreCOO, self).__init__(
            _path=path,
            _dim_cols=dim_cols,
            _value_col=value_col,
            _dtype=dtype,
            _storage_options=storage_options,
            _global_index=global_index,
            _sparse=sparse,
            **kw
        )

    @property
    def input(self):
        return self._input

    @property
    def path(self):
        return self._path

    @property
    def dim_cols(self):
        return self._dim_cols

    @property
    def value_col(self):
        return self._value_col

    @property
    def storage_options(self):
        return self._storage_options

    @property
    def global_index(self):
        return self._global_index

    @property
    def axis_offsets(self):
        return self._axis_offsets

    def _set_inputs(self, inputs):
        super(TensorStoreCOO, self)._set_inputs(inputs)
        self._input = self._inputs[0]

    def calc_shape(self, *inputs_shape):
        return (0,) * len(inputs_shape[0])

    @classmethod
    def tile(cls, op):
        in_tensor = op.input

        out_chunks = []
        out_chunk_shape = (0,) * in_tensor.ndim
        axis_offsets = [[0] + np.cumsum(ns)[:-1].tolist() for ns in in_tensor.nsplits]
        for chunk in in_tensor.chunks:
            chunk_op = op.copy().reset_key()
            chunk_path = "%s/%s.parquet" % (
                chunk_op.path,
                ",".join(str(j) for j in chunk.index),
            )
            chunk_op._path = chunk_path
            chunk_op._axis_offsets = tuple(
                axis_offsets[axis][idx] for axis, idx in enumerate(chunk.index)
            )
            out_chunk = chunk_op.new_chunk(
                [chunk], shape=out_chunk_shape, index=chunk.index
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs,
            op.outputs[0].shape,
            chunks=out_chunks,
            nsplits=((0,) * len(ns) for ns in in_tensor.nsplits),
        )

    @classmethod
    def execute(cls, ctx, op):
        import numpy as np
        import pandas as pd
        import pyarrow as pa
        import pyarrow.parquet as pq
        from ..io import open as fs_open

        to_store_data = ctx[op.inputs[0].key]
        storage_opts = json.loads(op.storage_options)
        axis_offsets = op.axis_offsets
        store_global_index = op.global_index
        dim_cols = op.dim_cols
        col_to_array = {}

        if isinstance(to_store_data, SparseNDArray):
            # sparse, convert to coo matrix
            matrix = to_store_data.raw.tocoo(copy=False)
            ndim = matrix.ndim

            if len(dim_cols) > 1:
                col_to_array[dim_cols[0]] = matrix.row
                if store_global_index:
                    # global index
                    col_to_array["global_" + dim_cols[0]] = matrix.row + axis_offsets[0]
                col_to_array[dim_cols[1]] = matrix.col
                if store_global_index:
                    col_to_array["global_" + dim_cols[1]] = matrix.col + axis_offsets[1]
            else:
                col_to_array[dim_cols[0]] = matrix.col
                if store_global_index:
                    col_to_array["global_" + dim_cols[0]] = matrix.col + axis_offsets[0]

            col_to_array[op.value_col] = matrix.data
        else:
            # dense, convert to numpy array
            arr = as_np_array(to_store_data)
            ndim = arr.ndim

            index = (
                np.array(np.meshgrid(*[np.arange(s) for s in arr.shape]))
                .T.reshape(-1, arr.ndim)
                .T
            )
            for j, col, ind in zip(range(len(dim_cols)), dim_cols, index):
                col_to_array[col] = ind
                if store_global_index:
                    col_to_array["global_" + col] = ind + axis_offsets[j]
            col_to_array[op.value_col] = arr.ravel()

        df = pd.DataFrame(col_to_array)
        if len(op.dim_cols) > ndim:
            for col in op.dim_cols[ndim:]:
                df[col] = None
        table = pa.Table.from_pandas(df)
        bio = BytesIO()
        pq.write_table(table, bio)
        bio.seek(0)

        # write oss
        with fs_open(op.path, "wb", **storage_opts) as out_file:
            out_file.write(bio.read())

        ctx[op.outputs[0].key] = np.empty((0,) * to_store_data.ndim)


def write_coo(a, path, dim_cols, value_col, **storage_opts):
    if a.ndim > len(dim_cols):
        raise ValueError(
            "dim_cols must have the same size as or greater than tensor's ndim"
        )

    global_index = storage_opts.pop("global_index", True)
    op = TensorStoreCOO(
        dtype=a.dtype,
        path=path,
        dim_cols=dim_cols,
        value_col=value_col,
        storage_options=json.dumps(storage_opts),
        sparse=a.issparse(),
        global_index=global_index,
    )
    return op(a)



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import os
import socket
import time
from urllib.parse import urlparse

import mars.oscar as mo
from mars.deploy.kubernetes.supervisor import K8SSupervisorCommandRunner
from mars.services.session import SessionAPI
from mars.services.web import OscarWebAPI

from ....config import options
from ..cupid_service import CupidServiceClient
from .core import CupidCommandRunnerMixin

DEFAULT_MARS_IDLE_TIMEOUT = 3 * 3600
CUPID_LAST_IDLE_TIME_KEY = "MarsServiceLastIdleTime"

logger = logging.getLogger(__name__)


class ClusterIdleCheckActor(mo.Actor):
    async def __post_create__(self):
        if "MARS_INSTANCE_IDLE_TIMEOUT" in os.environ:
            self._idle_timeout = int(os.environ["MARS_INSTANCE_IDLE_TIMEOUT"])
        else:
            self._idle_timeout = DEFAULT_MARS_IDLE_TIMEOUT
        self._idle_check_task = asyncio.create_task(self._check_idle())

    async def __pre_destroy__(self):
        self._idle_check_task.cancel()

    async def _check_idle(self):
        try:
            cupid_client = CupidServiceClient()
            session_api = await SessionAPI.create(self.address)

            last_idle_time_str = cupid_client.get_kv(CUPID_LAST_IDLE_TIME_KEY) or 0
            last_idle_time_from_service = await session_api.get_last_idle_time(None)
            if not last_idle_time_str:
                last_idle_time = last_idle_time_from_service or time.time()
            else:
                last_idle_time = float(last_idle_time_str)
        except:
            logger.exception("Failed to check instance idle")
            raise

        has_sessions_ever = False
        while True:
            try:
                await asyncio.sleep(10)
                idle_time_from_service = await session_api.get_last_idle_time(None)
                idle_time_from_service = idle_time_from_service or time.time()
                if idle_time_from_service != last_idle_time_from_service:
                    last_idle_time = idle_time_from_service
                    last_idle_time_from_service = idle_time_from_service

                if not has_sessions_ever:
                    # when no sessions has been created ever, we shall never stop our cluster
                    last_idle_time = time.time()
                    has_sessions_ever = len(await session_api.get_sessions()) > 0

                if time.time() - last_idle_time > self._idle_timeout:
                    logger.warning(
                        "Stopping instance due to idle timed out. %s > %s",
                        time.time() - last_idle_time,
                        self._idle_timeout,
                    )
                    await asyncio.to_thread(self._sync_stop_instance, cupid_client)
                else:
                    # make sure elapsed time is persisted
                    await asyncio.to_thread(
                        cupid_client.put_kv,
                        CUPID_LAST_IDLE_TIME_KEY,
                        str(last_idle_time),
                    )
            except:
                logger.exception("Failed to check instance idle")

    @staticmethod
    def _sync_stop_instance(cupid_client):
        from cupid import ContainerStatus, WorkItemProgress

        cupid_client.report_container_status(
            ContainerStatus.TERMINATED,
            "Instance idle timed out, stopping",
            WorkItemProgress.WIP_TERMINATING,
        )
        time.sleep(options.mars.container_status_timeout)
        # when instance is still not stopped, we kill forcifully
        cupid_client.terminate_instance(os.environ["MARS_K8S_POD_NAMESPACE"])


class CupidSupervisorCommandRunner(CupidCommandRunnerMixin, K8SSupervisorCommandRunner):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._register_task = None

    def __call__(self, *args, **kwargs):
        try:
            self.fix_protobuf_import()
            self.start_cupid_service()
            super().__call__(*args, **kwargs)
        except:
            import traceback

            traceback.print_exc()
            raise
        finally:
            self.stop_cupid_service()

    def parse_args(self, parser, argv, environ=None):
        self.fix_hyper_address()
        return super().parse_args(parser, argv, environ)

    async def _repeat_register(self, internal_endpoint, app_endpoint):
        from .client import CUPID_APP_NAME

        while True:
            try:
                await self.register_application(
                    CUPID_APP_NAME, internal_endpoint, app_endpoint
                )
                await asyncio.sleep(5)
            except asyncio.asyncio.CancelledError:
                break

    async def _register_web_endpoint(self):
        web_api = await OscarWebAPI.create(self.args.endpoint)
        web_port = urlparse(await web_api.get_web_address()).port

        host_addr = socket.gethostbyname(socket.gethostname())
        internal_endpoint = "http://{0}:{1}".format(host_addr, web_port)

        if os.environ.get("VM_ENGINE_TYPE") == "hyper":
            app_endpoint = socket.gethostname() + "-{}".format(web_port)
        else:
            app_endpoint = internal_endpoint

        self._register_task = asyncio.create_task(
            self._repeat_register(internal_endpoint, app_endpoint)
        )

    async def start_services(self):
        await self.write_node_endpoint()
        await super().start_services()
        await mo.create_actor(
            ClusterIdleCheckActor,
            uid=ClusterIdleCheckActor.default_uid(),
            address=self.args.endpoint,
        )
        await self._register_web_endpoint()

    async def stop_services(self):
        if self._register_task is not None:
            self._register_task.cancel()
        return await super().stop_services()


main = CupidSupervisorCommandRunner()

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import operator
import itertools

from ..core import Backend
from ...expr.expressions import ProjectCollectionExpr, FilterCollectionExpr, \
    LateralViewCollectionExpr, Column, SequenceExpr
from ...expr.element import IsIn, Func, MappedExpr
from ...expr.merge import InnerJoin
from ...expr.arithmetic import And
from ...expr.reduction import SequenceReduction
from ...expr.window import Window
from ...expr.merge import UnionCollectionExpr
from ...utils import traverse_until_source
from .utils import change_input, copy_sequence
from ....compat import reduce


class PredicatePushdown(Backend):
    def __init__(self, dag):
        self._dag = dag
        self._traversed = set()

    def pushdown(self):
        for node in traverse_until_source(self._dag, top_down=True, traversed=self._traversed):
            try:
                node.accept(self)
            except NotImplementedError:
                continue

        return self._dag.root

    def visit_filter_collection(self, expr):
        from .core import Optimizer

        predicates = self._split_filter_into_predicates(expr)

        if any(isinstance(seq, (Window, SequenceReduction, Func))
               for seq in itertools.chain(*expr.all_path(expr.input, strict=True))):
            # if any reduction like sum or window op exists, skip
            return
        if any(isinstance(p, IsIn) and
                (p._values is None or
                 (len(p._values) == 1 and isinstance(p._values[0], SequenceExpr)))
               for p in predicates):
            return

        filters = [expr]
        input = expr.input
        while isinstance(input, FilterCollectionExpr):
            filters.append(input)
            input = input.input

        def sub():
            if len(filters) > 1:
                new_expr = Optimizer.get_compact_filters(self._dag, *filters)
                old_predicate = expr._predicate
                expr._predicate = new_expr.predicate
                self._dag.substitute(old_predicate, expr._predicate, parents=[expr])
                [self._traversed.add(id(i)) for i in filters[1:]]
                return True
        if isinstance(input, ProjectCollectionExpr):
            if any(isinstance(seq, (Window, SequenceReduction, MappedExpr))
                   for seq in input.traverse(top_down=True, unique=True,
                                             stop_cond=lambda x: x is input.input)):
                # if any reduction like sum or window op exists, skip
                return
            if sub():
                predicates = self._split_filter_into_predicates(expr)
            predicate = expr.predicate
            remain = None
            if isinstance(input, LateralViewCollectionExpr):
                udtf_columns = set()
                for lv in input.lateral_views:
                    udtf_columns.update(lv.schema.names)
                pushes = []
                remains = []
                for p in predicates:
                    cols = [col.source_name for path in p.all_path(input, strict=True)
                            for col in path if isinstance(col, Column)]
                    if not (set(cols) & udtf_columns):
                        pushes.append(p)
                    else:
                        remains.append(p)
                if not pushes:
                    return
                predicate = reduce(operator.and_, pushes)
                if remains:
                    remain = reduce(operator.and_, remains)
            self._push_down_through_projection(predicate, expr, input)
            if remain is None:
                self._dag.substitute(expr, expr.input)
            else:
                expr.substitute(expr._predicate, remain, dag=self._dag)
        elif isinstance(input, InnerJoin):
            if sub():
                predicates = self._split_filter_into_predicates(expr)
            remains = []
            predicates = predicates[::-1]
            for i, predicate in enumerate(predicates):
                collection = self._predicate_on_same_collection(
                    predicate, expr.input, input)
                if collection is not False:
                    self._push_down_through_join(predicate, expr, input, collection)
                else:
                    remains.append(i)
            if len(remains) == 0:
                self._dag.substitute(expr, expr.input)
            else:
                expr.substitute(expr._predicate,
                                reduce(operator.and_, [predicates[i] for i in remains]),
                                dag=self._dag)
        elif isinstance(input, UnionCollectionExpr):
            sub()
            self._push_down_through_union(expr.predicate, expr, input)
            self._dag.substitute(expr, expr.input)

    def visit_join(self, expr):
        if not isinstance(expr, InnerJoin):
            return

        predicates = expr._predicate
        if predicates is None:
            return

        splits = self._split_predicates(predicates)
        rests = []

        left_filters, right_filters = [], []
        for i, p in enumerate(splits):
            collection = None
            skip = False
            for node in (expr.lhs, expr.rhs):
                if p.is_ancestor(node):
                    if collection is not None:
                        skip = True
                        break
                    collection = node
            if skip:
                rests.append(i)
                continue

            if collection is expr.lhs:
                left_filters.append(p)
            else:
                right_filters.append(p)

        if left_filters:
            left_filters = [copy_sequence(f, expr.lhs) for f in left_filters]
            new_lhs = expr.lhs.filter(*left_filters)
            self._dag.substitute(expr.lhs, new_lhs)
        if right_filters:
            right_filters = [copy_sequence(f, expr.rhs) for f in right_filters]
            new_rhs = expr.rhs.filter(*right_filters)
            self._dag.substitute(expr.rhs, new_rhs)

        if len(rests) != len(splits):
            new_predicates = [splits[i] for i in rests]
            join = expr.copy()
            join._predicate = new_predicates
            self._dag.substitute(expr, join)

    @classmethod
    def _split_predicates(cls, predicats):
        splits = [p for p in predicats if not isinstance(p, And)]
        nodes = list(p for p in predicats if isinstance(p, And))
        while len(nodes) > 0:
            node = nodes.pop(0)
            for s in (node.lhs, node.rhs):
                if not isinstance(s, And):
                    splits.append(s)
                else:
                    nodes.append(s)

        return splits

    @classmethod
    def _split_filter_into_predicates(cls, expr):
        predicate = expr.predicate
        if not isinstance(predicate, And):
            return [predicate]

        return cls._split_predicates([predicate])

    @classmethod
    def _predicate_on_same_collection(cls, predicate, input, joined):
        origins = set()

        cols = [col.source_name for path in predicate.all_path(input, strict=True)
                for col in path if isinstance(col, Column)]
        for col in cols:
            origins.add(joined._column_origins[col][0])
            if len(origins) > 1:
                return False

        # todo maybe broadcast in every collection in future
        if len(origins) == 0:
            return False

        return (joined.lhs, joined.rhs)[origins.pop()]

    def _push_down_through_projection(self, predicate, expr, projection):
        def get_project_field(n, col):
            field = projection._fields[projection.schema._name_indexes[col]]
            return copy_sequence(field, projection.input, self._dag)
        if isinstance(predicate, Column) and predicate.input is expr.input:
            # filter on column
            predicate = get_project_field(projection.input, predicate.source_name)
            if predicate.is_renamed():
                predicate = predicate.rename(predicate.name)
        else:
            change_input(predicate, expr.input, projection.input, get_project_field, self._dag)

        filter_expr = FilterCollectionExpr(
            projection.input, predicate, _schema=projection.input.schema)
        get_filter_field = lambda filter_node, col: filter_node[col]
        change_input(projection, projection.input, filter_expr, get_filter_field, self._dag)

    def _push_down_through_join(self, predicate, expr, joined, collection):
        def get_join_field(node, col):
            return node[joined._column_origins[col][1]]

        change_input(predicate, expr.input, collection, get_join_field, self._dag)

        filter_expr = FilterCollectionExpr(
            collection, predicate, _schema=collection.schema)
        get_field = lambda n, col: n[col]
        change_input(joined, collection, filter_expr, get_field, self._dag)

    def _push_down_through_union(self, predicate, expr, union):
        get_field = lambda n, col: n[col]

        predicates = [copy_sequence(predicate, expr.input, self._dag) for _ in range(2)]
        for predicate, collection in zip(predicates, (union.lhs, union.rhs)):
            change_input(predicate, expr.input, collection, get_field, self._dag)

            filter_expr = FilterCollectionExpr(collection, predicate, _schema=collection.schema)
            change_input(union, collection, filter_expr, get_field, self._dag)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import time
import uuid
import logging

import requests
from typing import List
from mars.core.context import get_context
from mars.oscar.errors import ActorNotExist
from mars.core import OutputType
from mars.dataframe.operands import DataFrameOperandMixin, DataFrameOperand
from mars.dataframe.utils import build_concatenated_rows_frame, parse_index
from mars.serialization.serializables import (
    StringField,
    SeriesField,
    BoolField,
    DictField,
    Int64Field,
    ListField,
    FieldTypes,
)

from ....config import options
from ....utils import to_str
from ..cupid_service import CupidServiceClient

logger = logging.getLogger(__name__)

_output_type_kw = dict(_output_types=[OutputType.dataframe])


class DataFrameWriteTable(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = 123460

    dtypes = SeriesField("dtypes")

    odps_params = DictField("odps_params", default=None)
    table_name = StringField("table_name", default=None)
    partition_spec = StringField("partition_spec", default=None)
    partition_columns = ListField("partition_columns", FieldTypes.string, default=None)
    overwrite = BoolField("overwrite", default=None)
    write_batch_size = Int64Field("write_batch_size", default=None)
    unknown_as_string = BoolField("unknown_as_string", default=None)
    tunnel_quota_name = StringField("tunnel_quota_name", default=None)

    def __init__(self, **kw):
        kw.update(_output_type_kw)
        super(DataFrameWriteTable, self).__init__(**kw)

    @property
    def retryable(self):
        return "CUPID_SERVICE_SOCKET" not in os.environ

    def __call__(self, x):
        shape = (0,) * len(x.shape)
        index_value = parse_index(x.index_value.to_pandas()[:0], x.key, "index")
        columns_value = parse_index(
            x.columns_value.to_pandas()[:0], x.key, "columns", store_data=True
        )
        return self.new_dataframe(
            [x],
            shape=shape,
            dtypes=x.dtypes[:0],
            index_value=index_value,
            columns_value=columns_value,
        )

    @classmethod
    def _tile_cupid(cls, op):
        from mars.dataframe.utils import build_concatenated_rows_frame

        cupid_client = CupidServiceClient()
        upload_handle = cupid_client.create_table_upload_session(
            op.odps_params, op.table_name
        )

        input_df = build_concatenated_rows_frame(op.inputs[0])
        out_df = op.outputs[0]

        out_chunks = []
        out_chunk_shape = (0,) * len(input_df.shape)
        blocks = {}
        for chunk in input_df.chunks:
            block_id = str(int(time.time())) + "_" + str(uuid.uuid4()).replace("-", "")
            chunk_op = DataFrameWriteTableSplit(
                dtypes=op.dtypes,
                table_name=op.table_name,
                odps_params=op.odps_params,
                unknown_as_string=op.unknown_as_string,
                partition_spec=op.partition_spec,
                cupid_handle=to_str(upload_handle),
                block_id=block_id,
                write_batch_size=op.write_batch_size,
            )
            out_chunk = chunk_op.new_chunk(
                [chunk],
                shape=out_chunk_shape,
                index=chunk.index,
                index_value=out_df.index_value,
                dtypes=chunk.dtypes,
            )
            out_chunks.append(out_chunk)
            blocks[block_id] = op.partition_spec

        # build commit tree
        combine_size = 8
        chunks = out_chunks
        while len(chunks) >= combine_size:
            new_chunks = []
            for i in range(0, len(chunks), combine_size):
                chks = chunks[i : i + combine_size]
                if len(chks) == 1:
                    chk = chks[0]
                else:
                    chk_op = DataFrameWriteTableCommit(
                        dtypes=op.dtypes, is_terminal=False
                    )
                    chk = chk_op.new_chunk(
                        chks,
                        shape=out_chunk_shape,
                        index_value=out_df.index_value,
                        dtypes=op.dtypes,
                    )
                new_chunks.append(chk)
            chunks = new_chunks

        assert len(chunks) < combine_size

        commit_table_op = DataFrameWriteTableCommit(
            dtypes=op.dtypes,
            table_name=op.table_name,
            blocks=blocks,
            cupid_handle=to_str(upload_handle),
            overwrite=op.overwrite,
            odps_params=op.odps_params,
            is_terminal=True,
        )
        commit_table_chunk = commit_table_op.new_chunk(
            chunks,
            shape=out_chunk_shape,
            dtypes=op.dtypes,
            index_value=out_df.index_value,
            index=(0,) * len(out_chunk_shape),
        )

        new_op = op.copy()
        return new_op.new_dataframes(
            op.inputs,
            shape=out_df.shape,
            index_value=out_df.index_value,
            dtypes=out_df.dtypes,
            columns_value=out_df.columns_value,
            chunks=[commit_table_chunk],
            nsplits=((0,),) * len(out_chunk_shape),
        )

    @classmethod
    def _tile_tunnel(cls, op):
        from odps import ODPS

        out_df = op.outputs[0]

        if op.overwrite:
            o = ODPS(
                op.odps_params["access_id"],
                op.odps_params["secret_access_key"],
                project=op.odps_params["project"],
                endpoint=op.odps_params["endpoint"],
            )
            data_target = o.get_table(op.table_name)
            if op.partition_spec:
                data_target = data_target.get_partition(op.partition_spec)
            data_target.truncate()

        in_df = build_concatenated_rows_frame(op.inputs[0])
        logger.info("Tile table %s[%s]", op.table_name, op.partition_spec)
        recorder_name = str(uuid.uuid4())
        out_chunks = []
        for chunk in in_df.chunks:
            chunk_op = DataFrameWriteTableSplit(
                dtypes=op.dtypes,
                table_name=op.table_name,
                odps_params=op.odps_params,
                partition_spec=op.partition_spec,
                commit_recorder_name=recorder_name,
                tunnel_quota_name=op.tunnel_quota_name,
            )
            index_value = parse_index(chunk.index_value.to_pandas()[:0], chunk)
            out_chunk = chunk_op.new_chunk(
                [chunk],
                shape=(0, 0),
                index_value=index_value,
                columns_value=out_df.columns_value,
                dtypes=out_df.dtypes,
                index=chunk.index,
            )
            out_chunks.append(out_chunk)
        ctx = get_context()
        ctx.create_remote_object(recorder_name, _TunnelCommitRecorder, len(out_chunks))
        new_op = op.copy()
        params = out_df.params.copy()
        params.update(
            dict(chunks=out_chunks, nsplits=((0,) * in_df.chunk_shape[0], (0,)))
        )
        return new_op.new_tileables([in_df], **params)

    @classmethod
    def tile(cls, op):
        if "CUPID_SERVICE_SOCKET" in os.environ:
            return cls._tile_cupid(op)
        else:
            return cls._tile_tunnel(op)


class DataFrameWriteTableSplit(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = 123461

    dtypes = SeriesField("dtypes")
    table_name = StringField("table_name")
    partition_spec = StringField("partition_spec", default=None)
    cupid_handle = StringField("cupid_handle", default=None)
    block_id = StringField("block_id", default=None)
    write_batch_size = Int64Field("write_batch_size", default=None)
    unknown_as_string = BoolField("unknown_as_string", default=None)
    commit_recorder_name = StringField("commit_recorder_name", default=None)

    # for tunnel
    odps_params = DictField("odps_params", default=None)
    tunnel_quota_name = StringField("tunnel_quota_name", default=None)

    def __init__(self, **kw):
        kw.update(_output_type_kw)
        super(DataFrameWriteTableSplit, self).__init__(**kw)

    @property
    def retryable(self):
        return "CUPID_SERVICE_SOCKET" not in os.environ

    @classmethod
    def _execute_in_cupid(cls, ctx, op):
        import os

        import pandas as pd
        from odps import ODPS
        from odps.accounts import BearerTokenAccount

        cupid_client = CupidServiceClient()
        to_store_data = ctx[op.inputs[0].key]

        bearer_token = cupid_client.get_bearer_token()
        account = BearerTokenAccount(bearer_token)
        project = os.environ.get("ODPS_PROJECT_NAME", None)
        odps_params = op.odps_params.copy()
        if project:
            odps_params["project"] = project
        endpoint = os.environ.get("ODPS_RUNTIME_ENDPOINT") or odps_params["endpoint"]
        o = ODPS(
            None,
            None,
            account=account,
            project=odps_params["project"],
            endpoint=endpoint,
        )
        table_obj = o.get_table(op.table_name)
        odps_schema = table_obj.table_schema
        project_name = table_obj.project.name
        schema_name = table_obj.get_schema().name if table_obj.get_schema() is not None else None
        table_name = table_obj.name

        writer_config = dict(
            _table_name=table_name,
            _project_name=project_name,
            _schema_name=schema_name,
            _table_schema=odps_schema,
            _partition_spec=op.partition_spec,
            _block_id=op.block_id,
            _handle=op.cupid_handle,
        )
        cupid_client.write_table_data(writer_config, to_store_data, op.write_batch_size)
        ctx[op.outputs[0].key] = pd.DataFrame()

    @classmethod
    def _execute_arrow_tunnel(cls, ctx, op):
        from odps import ODPS
        from odps.tunnel import TableTunnel
        import pyarrow as pa
        import pandas as pd

        project = os.environ.get("ODPS_PROJECT_NAME", None)
        odps_params = op.odps_params.copy()
        if project:
            odps_params["project"] = project
        endpoint = os.environ.get("ODPS_RUNTIME_ENDPOINT") or odps_params["endpoint"]
        o = ODPS(
            odps_params["access_id"],
            odps_params["secret_access_key"],
            project=odps_params["project"],
            endpoint=endpoint,
        )

        t = o.get_table(op.table_name)
        tunnel = TableTunnel(o, project=t.project, quota_name=op.tunnel_quota_name)
        retry_times = options.retry_times
        init_sleep_secs = 1
        split_index = op.inputs[0].index
        logger.info(
            "Start creating upload session for table %s split index %s retry_times %s.",
            op.table_name,
            split_index,
            retry_times,
        )
        retries = 0
        while True:
            try:
                if op.partition_spec is not None:
                    upload_session = tunnel.create_upload_session(
                        t.name, partition_spec=op.partition_spec
                    )
                else:
                    upload_session = tunnel.create_upload_session(t.name)
                break
            except:
                if retries >= retry_times:
                    raise
                retries += 1
                sleep_secs = retries * init_sleep_secs
                logger.exception(
                    "Create upload session failed, sleep %s seconds and retry it",
                    sleep_secs,
                    exc_info=1,
                )
                time.sleep(sleep_secs)
        logger.info(
            "Start writing table %s. split_index: %s tunnel_session: %s",
            op.table_name,
            split_index,
            upload_session.id,
        )
        retries = 0
        while True:
            try:
                writer = upload_session.open_arrow_writer(0)
                arrow_rb = pa.RecordBatch.from_pandas(ctx[op.inputs[0].key])
                writer.write(arrow_rb)
                writer.close()
                break
            except:
                if retries >= retry_times:
                    raise
                retries += 1
                sleep_secs = retries * init_sleep_secs
                logger.exception(
                    "Write data failed, sleep %s seconds and retry it",
                    sleep_secs,
                    exc_info=1,
                )
                time.sleep(sleep_secs)
        recorder_name = op.commit_recorder_name
        try:
            recorder = ctx.get_remote_object(recorder_name)
        except ActorNotExist:
            while True:
                logger.info(
                    "Writing table %s has been finished, waiting to be canceled by speculative scheduler",
                    op.table_name,
                )
                time.sleep(3)
        can_commit, can_destroy = recorder.try_commit(split_index)
        if can_commit:
            # FIXME If this commit failed or the process crashed, the whole write will still raise error.
            # But this situation is very rare so we skip the error handling.
            logger.info(
                "Committing to table %s with upload session %s", op.table_name, upload_session.id
            )
            upload_session.commit([0])
            logger.info(
                "Finish writing table %s. split_index: %s tunnel_session: %s",
                op.table_name,
                split_index,
                upload_session.id,
            )
        else:
            logger.info(
                "Skip writing table %s. split_index: %s", op.table_name, split_index
            )
        if can_destroy:
            try:
                ctx.destroy_remote_object(recorder_name)
                logger.info("Delete remote object %s", recorder_name)
            except ActorNotExist:
                pass
        logger.info(
            "Committing to table %s with upload session %s", op.table_name, upload_session.id
        )
        upload_session.commit([0])
        logger.info(
            "Finish writing table %s. split_index: %s tunnel_session: %s",
            op.table_name,
            split_index,
            upload_session.id,
        )
        ctx[op.outputs[0].key] = pd.DataFrame()

    @classmethod
    def execute(cls, ctx, op):
        if op.cupid_handle is not None:
            cls._execute_in_cupid(ctx, op)
        else:
            cls._execute_arrow_tunnel(ctx, op)


class _TunnelCommitRecorder:
    _commit_status: List[bool]

    def __init__(self, n_chunk: int):
        self._n_chunk = n_chunk
        self._commit_status = {}

    def try_commit(self, index: tuple):
        if index in self._commit_status:
            return False, len(self._commit_status) == self._n_chunk
        else:
            self._commit_status[index] = True
            return True, len(self._commit_status) == self._n_chunk


class DataFrameWriteTableCommit(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = 123462

    dtypes = SeriesField("dtypes")

    odps_params = DictField("odps_params")
    table_name = StringField("table_name")
    overwrite = BoolField("overwrite", default=False)
    blocks = DictField("blocks", default=None)
    cupid_handle = StringField("cupid_handle", default=None)
    is_terminal = BoolField("is_terminal", default=None)

    def __init__(self, **kw):
        kw.update(_output_type_kw)
        super(DataFrameWriteTableCommit, self).__init__(**kw)

    @classmethod
    def execute(cls, ctx, op):
        import pandas as pd
        from ..cupid_service import CupidServiceClient

        if op.is_terminal:
            odps_params = op.odps_params.copy()
            project = os.environ.get("ODPS_PROJECT_NAME", None)
            if project:
                odps_params["project"] = project

            client = CupidServiceClient()
            client.commit_table_upload_session(
                odps_params, op.table_name, op.cupid_handle, op.blocks, op.overwrite
            )

        ctx[op.outputs[0].key] = pd.DataFrame()


def write_odps_table(
    df,
    table,
    partition=None,
    overwrite=False,
    unknown_as_string=None,
    odps_params=None,
    write_batch_size=None,
    tunnel_quota_name=None,
):
    op = DataFrameWriteTable(
        dtypes=df.dtypes,
        odps_params=odps_params,
        table_name=table.full_table_name,
        unknown_as_string=unknown_as_string,
        partition_spec=partition,
        overwrite=overwrite,
        write_batch_size=write_batch_size,
        tunnel_quota_name=tunnel_quota_name,
    )
    return op(df)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import sys
import tokenize

import numpy as np
import pandas as pd

from ..compat import StringIO
from ..df.backends.pd.types import _np_to_df_types
from ..df.backends.odpssql.types import df_type_to_odps_type
from ..df import types
from ..errors import ODPSError


def pd_type_to_odps_type(dtype, col_name, unknown_as_string=None, project=None):
    import numpy as np

    if dtype in _np_to_df_types:
        df_type = _np_to_df_types[dtype]
    elif dtype == np.datetime64(0, "ns"):
        df_type = types.timestamp
    elif unknown_as_string:
        df_type = types.string
    else:
        raise ValueError(
            "Unknown type {}, column name is {},"
            "specify `unknown_as_string=True` "
            "or `as_type` to set column dtype".format(dtype, col_name)
        )

    return df_type_to_odps_type(df_type, project=project)


def use_odps2_type(func):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        from odps import options

        old_value = options.sql.use_odps2_extension
        options.sql.use_odps2_extension = True

        old_settings = options.sql.settings
        options.sql.settings = old_settings or {}
        options.sql.settings.update({"odps.sql.hive.compatible": True})
        options.sql.settings.update({"odps.sql.decimal.odps2": True})
        try:
            func(*args, **kwargs)
        finally:
            options.sql.use_odps2_extension = old_value
            options.sql.settings = old_settings

    return wrapped


def convert_pandas_object_to_string(df):
    import pandas as pd
    import numpy as np

    def convert_to_string(v):
        if v is None or pd.isna(v):
            return None
        else:
            return str(v)

    object_columns = [c for c, t in df.dtypes.iteritems() if t == np.dtype("O")]
    for col in object_columns:
        df[col] = df[col].map(convert_to_string)
    return df


def check_partition_exist(table, partition_spec):
    try:
        return table.exist_partition(partition_spec)
    except (ValueError, ODPSError):
        return False


def rewrite_partition_predicate(predicate, cols):
    cols = [c.lower() for c in cols]

    def _tokenize_predicate():
        reader = StringIO(predicate).readline
        last_num = ""
        last_bool_cmp = ""
        last_toknum = None

        for toknum, tokval, _, _, _ in tokenize.generate_tokens(reader):
            if last_toknum != toknum:
                # handle consecutive operands / numbers
                if last_num:
                    yield tokenize.STRING, '"%s"' % last_num
                    last_num = ""
                if last_bool_cmp:
                    if last_bool_cmp not in ("&", "&&", "|", "||"):
                        raise SyntaxError("Operand %s not recognized" % last_bool_cmp)

                    yield tokenize.OP, last_bool_cmp[-1]
                    last_bool_cmp = ""

            last_toknum = toknum

            if toknum == tokenize.NUMBER:
                last_num += tokval
            elif toknum == tokenize.OP and tokval in ("&", "|"):
                last_bool_cmp += tokval
            else:
                if toknum == tokenize.NAME:
                    lower_tokval = tokval.lower()
                    if lower_tokval == "max_pt":
                        tokval = "@" + lower_tokval
                    elif lower_tokval not in cols:
                        toknum = tokenize.STRING
                        tokval = '"%s"' % tokval
                elif toknum == tokenize.OP:
                    if tokval == "=":
                        tokval = "=="
                    elif tokval == ",":
                        tokval = "&"
                yield toknum, tokval

    return tokenize.untokenize(list(_tokenize_predicate()))


def calc_max_partition(
    table_name=None, odps=None, query_table_name=None, result_cache=None
):
    result_cache = result_cache or dict()
    table_name = table_name or query_table_name

    table = odps.get_table(table_name)
    if table_name in result_cache:
        return result_cache[table.full_table_name]

    return list(table.get_max_partition(skip_empty=True).partition_spec.kv.values())[0]


def filter_partitions(odps, partitions, predicate):
    cols = partitions[0].partition_spec.keys
    part_df = pd.DataFrame(
        [part.partition_spec.kv.values() for part in partitions], columns=cols
    )
    part_df = part_df.astype(str)
    part_df["__pt_obj__"] = pd.Series(partitions, dtype=np.dtype("O"))

    predicate = rewrite_partition_predicate(predicate, cols)

    global_dict = globals().copy()
    query_table = partitions[0].parent.parent
    query_table_name = query_table.project.name + "." + query_table.name
    result_cache = dict()
    global_dict["max_pt"] = functools.partial(
        calc_max_partition,
        odps=odps,
        query_table_name=query_table_name,
        result_cache=result_cache,
    )

    part_df.query(predicate, engine="python", global_dict=global_dict, inplace=True)

    return part_df["__pt_obj__"].to_list()


def get_default_resource_files(names, project="public"):
    """
    Use Python version plus architecture to name packages.
    For instance, public.pymars-0.8.6-cp37-x86_64.zip
    """
    suffix = (
        "-cp"
        + "".join(str(x) for x in sys.version_info[:2])
        + "-x86_64.zip"  # todo fix for other architectures
    )
    return [project + "." + name + suffix for name in names]


def build_mars_image_name(mars_image=None, with_dist_suffix=True):
    from cupid.config import options
    from cupid.utils import build_image_name

    prefix = options.cupid.image_prefix
    if prefix is not None:
        dockerhub_address = prefix.split("/")[0]
    else:
        dockerhub_address = None

    mars_image = mars_image or "mars"
    if ":" not in mars_image:
        mars_image = build_image_name(mars_image, with_dist_suffix=with_dist_suffix)
    if dockerhub_address is not None and dockerhub_address not in mars_image:
        mars_image = prefix + mars_image
    return mars_image



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .core import LazyLoad
from .. import serializers, utils


class XFlow(LazyLoad):
    __slots__ = ("xml_source",)

    name = serializers.XMLNodeField("Name")
    owner = serializers.XMLNodeField("Owner")
    creation_time = serializers.XMLNodeField(
        "CreationTime", parse_callback=utils.parse_rfc822
    )
    last_modified_time = serializers.XMLNodeField(
        "LastModifiedTime", parse_callback=utils.parse_rfc822
    )

    def reload(self):
        url = self.resource()
        resp = self._client.get(url)

        self.xml_source = resp.content
        self.owner = resp.headers.get("x-odps-owner")
        self.creation_time = utils.parse_rfc822(
            resp.headers.get("x-odps-creation-time")
        )
        self.last_modified_time = utils.parse_rfc822(resp.headers.get("Last-Modified"))

    def update(self):
        return self._parent.update(self)

    def drop(self):
        return self._parent.delete(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from collections import OrderedDict

try:
    import sqlalchemy
    has_sqlalchemy = True
except ImportError:
    has_sqlalchemy = False

from ... import types
from ....compat import six
from ....models import TableSchema


_sqlalchemy_to_df_types = OrderedDict()
_df_to_sqlalchemy_types = OrderedDict()

if has_sqlalchemy:
    try:
        _sqlalchemy_to_df_types[sqlalchemy.CHAR] = types.int8
        _sqlalchemy_to_df_types[sqlalchemy.SmallInteger] = types.int16
        _sqlalchemy_to_df_types[sqlalchemy.BigInteger] = types.int64
        _sqlalchemy_to_df_types[sqlalchemy.Integer] = types.int32
        _sqlalchemy_to_df_types[sqlalchemy.Float] = types.float64
        _sqlalchemy_to_df_types[sqlalchemy.String] = types.string
        _sqlalchemy_to_df_types[sqlalchemy.Boolean] = types.boolean
        _sqlalchemy_to_df_types[sqlalchemy.DECIMAL] = types.decimal
        _sqlalchemy_to_df_types[sqlalchemy.DateTime] = types.datetime

        _df_to_sqlalchemy_types = dict((v, k) for k, v in six.iteritems(_sqlalchemy_to_df_types))
        _df_to_sqlalchemy_types[types.string] = sqlalchemy.Text  # we store text
    except AttributeError as ex:
        warnings.warn(
            "Your sqlalchemy does not have attribute %s. If you need to use it "
            "with PyODPS, please install a newer release." % str(ex),
            ImportWarning,
        )


def sqlalchemy_to_df_type(sqlalchemy_type):
    for sqlalchemy_type_cls, df_type in six.iteritems(_sqlalchemy_to_df_types):
        if isinstance(sqlalchemy_type, sqlalchemy_type_cls):
            return df_type

    raise ValueError(
        'Cannot convert SQLAlchemy type %s to dataframe type' % sqlalchemy_type)


def sqlalchemy_to_df_schema(sqlalchemy_columns):
    names, types = [], []
    for c in sqlalchemy_columns:
        names.append(c.name)
        types.append(sqlalchemy_to_df_type(c.type))

    return TableSchema.from_lists(names, types)


def df_type_to_sqlalchemy_type(df_type, engine=None):
    if isinstance(df_type, six.string_types):
        df_type = types.validate_data_type(df_type)

    if isinstance(df_type, types.Datetime):
        return sqlalchemy.DateTime(timezone=True)
    elif df_type == types.decimal:
        return sqlalchemy.DECIMAL(36, 18)

    if engine and engine.name == 'mysql' and df_type == types.float64:
        from sqlalchemy.dialects.mysql.types import DOUBLE, DECIMAL
        return DOUBLE()
    return _df_to_sqlalchemy_types[df_type]


def df_schema_to_sqlalchemy_columns(df_schema, ignorecase=False, engine=None):
    names = [col.name.lower() if ignorecase else col.name
             for col in df_schema._columns]
    types = [df_type_to_sqlalchemy_type(col.type, engine=engine) for col in df_schema._columns]

    return [sqlalchemy.Column(n, t) for n, t in zip(names, types)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import contextlib
import logging
import threading
import traceback
import warnings
from copy import deepcopy

from .compat import six

try:
    import contextvars
except ImportError:
    contextvars = None


DEFAULT_CHUNK_SIZE = 65536
DEFAULT_CONNECT_RETRY_TIMES = 4
DEFAULT_CONNECT_TIMEOUT = 120
DEFAULT_READ_TIMEOUT = 120
DEFAULT_POOL_CONNECTIONS = 10
DEFAULT_POOL_MAXSIZE = 10
DEFAULT_RETRY_DELAY = 0.1
_DEFAULT_REDIRECT_WARN = 'Option {source} has been replaced by {target} and might be removed in a future release.'


class OptionError(Exception):
    pass


class Redirection(object):
    def __init__(self, item, warn=None):
        self._items = item.split('.')
        self._warn = warn
        self._warned = True
        self._parent = None

    def bind(self, attr_dict):
        self._parent = attr_dict
        self.getvalue()
        self._warned = False

    def getvalue(self, silent=False):
        if not silent and self._warn and not self._warned:
            in_completer = any(1 for st in traceback.extract_stack()
                               if 'completer' in st[0].lower())
            if not in_completer:
                self._warned = True
                warnings.warn(self._warn)
        conf = self._parent.root
        for it in self._items:
            conf = getattr(conf, it)
        return conf

    def setvalue(self, value, silent=False):
        if not silent and self._warn and not self._warned:
            self._warned = True
            warnings.warn(self._warn)
        conf = self._parent.root
        for it in self._items[:-1]:
            conf = getattr(conf, it)
        setattr(conf, self._items[-1], value)


class PandasRedirection(Redirection):
    def __init__(self, *args, **kw):
        super(PandasRedirection, self).__init__(*args, **kw)
        self._val = None
        try:
            import pandas  # noqa: F401

            self._use_pd = True
        except ImportError:
            self._use_pd = False

    def getvalue(self, silent=False):
        if self._use_pd:
            import pandas as pd
            try:
                return pd.get_option('.'.join(self._items))
            except (KeyError, LookupError, AttributeError):
                self._use_pd = False
        else:
            return self._val

    def setvalue(self, value, silent=False):
        if self._use_pd:
            import pandas as pd
            key = '.'.join(self._items)
            if value != pd.get_option(key):
                pd.set_option(key, value)
        else:
            self._val = value


class AttributeDict(dict):
    def __init__(self, *args, **kwargs):
        self._inited = False
        self._parent = kwargs.pop('_parent', None)
        self._root = None
        super(AttributeDict, self).__init__(*args, **kwargs)
        self._inited = True

    @property
    def root(self):
        if self._root is not None:
            return self._root
        if self._parent is None:
            self._root = self
        else:
            self._root = self._parent.root
        return self._root

    def __getattr__(self, item):
        if item in self:
            val = self[item]
            if isinstance(val, AttributeDict):
                return val
            elif isinstance(val[0], Redirection):
                return val[0].getvalue()
            else:
                return val[0]
        return object.__getattribute__(self, item)

    def __dir__(self):
        return list(six.iterkeys(self))

    def register(self, key, value, validator=None):
        self[key] = value, validator
        if isinstance(value, Redirection):
            value.bind(self)

    def unregister(self, key):
        del self[key]

    def add_validator(self, key, validator):
        value, old_validator = self[key]
        validators = getattr(
            old_validator, "validators", [old_validator] if callable(old_validator) else []
        )
        validators.append(validator)
        self[key] = (value, all_validator(*validators))

    def _setattr(self, key, value, silent=False):
        if not silent and key not in self:
            raise OptionError("Cannot identify configuration name '%s'." % str(key))

        if not isinstance(value, AttributeDict):
            validate = None
            if key in self:
                val = self[key]
                validate = self[key][1]
                if validate is not None:
                    if not validate(value):
                        raise ValueError('Cannot set value %s' % value)
                if isinstance(val[0], Redirection):
                    val[0].setvalue(value)
                else:
                    self[key] = value, validate
            else:
                self[key] = value, validate
        else:
            self[key] = value

    def __setattr__(self, key, value):
        if key == '_inited':
            super(AttributeDict, self).__setattr__(key, value)
            return
        try:
            object.__getattribute__(self, key)
            super(AttributeDict, self).__setattr__(key, value)
            return
        except AttributeError:
            pass

        if not self._inited:
            super(AttributeDict, self).__setattr__(key, value)
        else:
            self._setattr(key, value)

    def loads(self, d):
        dispatches = collections.defaultdict(dict)
        for k, v in six.iteritems(d):
            if '.' in k:
                sk, rk = k.split('.', 1)
                dispatches[sk][rk] = v
            elif isinstance(self[k][0], Redirection):
                self[k][0].setvalue(v, silent=True)
            else:
                setattr(self, k, v)
        for k, v in six.iteritems(dispatches):
            self[k].loads(v)

    def dumps(self):
        from .accounts import BaseAccount

        result_dict = dict()
        for k, v in six.iteritems(self):
            if isinstance(v, AttributeDict):
                result_dict.update((k + '.' + sk, sv) for sk, sv in six.iteritems(v.dumps()))
            elif isinstance(v[0], BaseAccount) or callable(v[0]):
                # ignore accounts in config dumps
                result_dict[k] = None
            elif isinstance(v[0], Redirection):
                result_dict[k] = v[0].getvalue(silent=True)
            else:
                result_dict[k] = v[0]
        return result_dict


class Config(object):
    def __init__(self, config=None):
        self._config = config or AttributeDict()

    def __dir__(self):
        return list(six.iterkeys(self._config))

    def __getattr__(self, item):
        return getattr(self._config, item)

    def __setattr__(self, key, value):
        if key == '_config':
            object.__setattr__(self, key, value)
            return
        setattr(self._config, key, value)

    def register_option(self, option, value, validator=None):
        assert validator is None or callable(validator)
        splits = option.split('.')
        conf = self._config

        for name in splits[:-1]:
            config = conf.get(name)
            if config is None:
                val = AttributeDict(_parent=conf)
                conf[name] = val
                conf = val
            elif not isinstance(config, dict):
                raise AttributeError(
                    'Fail to set option: %s, conflict has encountered' % option)
            else:
                conf = config

        key = splits[-1]
        if conf.get(key) is not None:
            raise AttributeError(
                'Fail to set option: %s, option has been set' % option)

        conf.register(key, value, validator)

    def register_pandas(self, option, value=None, validator=None):
        redir = PandasRedirection(option)
        self.register_option(option, redir, validator=validator)
        if value is not None:
            redir.setvalue(value)

    def redirect_option(self, option, target, warn=_DEFAULT_REDIRECT_WARN):
        redir = Redirection(target, warn=warn.format(source=option, target=target))
        self.register_option(option, redir)

    def unregister_option(self, option):
        splits = option.split('.')
        conf = self._config
        for name in splits[:-1]:
            config = conf.get(name)
            if not isinstance(config, dict):
                raise AttributeError(
                    'Fail to unregister option: %s, conflict has encountered' % option)
            else:
                conf = config

        key = splits[-1]
        if key not in conf:
            raise AttributeError('Option %s not configured, thus failed to unregister.' % option)
        conf.unregister(key)

    def update(self, new_config):
        if not isinstance(new_config, dict):
            new_config = new_config._config
        for option, value in new_config.items():
            try:
                self.register_option(option, value)
            except AttributeError:
                setattr(self, option, value)

    def add_validator(self, option, validator):
        splits = option.split('.')
        conf = self._config
        for name in splits[:-1]:
            config = conf.get(name)
            if not isinstance(config, dict):
                raise AttributeError(
                    'Fail to add validator: %s, conflict has encountered' % option)
            else:
                conf = config

        key = splits[-1]
        if key not in conf:
            raise AttributeError('Option %s not configured, thus failed to set validator.' % option)
        conf.add_validator(key, validator)

    def loads(self, d):
        return self._config.loads(d)

    def dumps(self):
        return self._config.dumps()


def is_interactive():
    import __main__ as main
    return not hasattr(main, '__file__')


# validators
def any_validator(*validators):
    def validate(x):
        return any(validator(x) for validator in validators)
    return validate


def all_validator(*validators):
    def validate(x):
        return all(validator(x) for validator in validators)

    validate.validators = validators
    return validate


is_null = lambda x: x is None
is_bool = lambda x: isinstance(x, bool)
is_float = lambda x: isinstance(x, float)
is_integer = lambda x: isinstance(x, six.integer_types)
is_string = lambda x: isinstance(x, six.string_types)
is_dict = lambda x: isinstance(x, dict)


def is_in(vals):
    def validate(x):
        return x in vals
    return validate


def show_deprecated(msg):
    def validator(x):
        warnings.warn(msg, DeprecationWarning)
        return x

    return validator


def verbose_log_validator(val):
    if options.verbose == val:
        # no flip, return directly
        return True

    logging.basicConfig()
    odps_logger = logging.getLogger("odps")
    odps_logger.propagate = False
    if _stream_handler not in odps_logger.handlers:
        odps_logger.addHandler(_stream_handler)

    if val:
        logging.basicConfig()
        old_effective_level = odps_logger.getEffectiveLevel() or logging.WARNING
        target_level = min(logging.INFO, old_effective_level)

        odps_logger.oldLevel = odps_logger.level
        odps_logger.setLevel(target_level)

        _stream_handler.oldLevel = _stream_handler.level
        _stream_handler.setLevel(target_level)
        odps_logger.addHandler(_verbose_log_handler)
    else:
        old_level = getattr(odps_logger, "oldLevel", logging.NOTSET)
        odps_logger.setLevel(old_level)

        old_level = getattr(_stream_handler, "oldLevel", logging.NOTSET)
        _stream_handler.setLevel(old_level)
        odps_logger.removeHandler(_verbose_log_handler)
    return True


class VerboseLogHandler(logging.StreamHandler):
    def emit(self, record):
        try:
            if options.verbose_log:
                msg = self.format(record)
                options.verbose_log(msg)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)


_stream_handler = logging.StreamHandler()
_verbose_log_handler = VerboseLogHandler()


default_options = Config()
default_options.register_option('is_global_account_overwritable', True, validator=is_bool)
default_options.register_option('account', None)
default_options.register_option('endpoint', None)
default_options.redirect_option('end_point', 'endpoint')
default_options.register_option('default_project', None)
default_options.register_option('default_schema', None)
default_options.register_option('app_account', None)
default_options.register_option('region_name', None)
default_options.register_option('local_timezone', None)
default_options.register_option('use_legacy_parsedate', False)
default_options.register_option('allow_antique_date', False)
default_options.register_option(
    'user_agent_pattern', '$pyodps_version $python_version $os_version $maxframe_version'
)
default_options.register_option('logview_host', None)
default_options.register_option('logview_hours', 24 * 30, validator=is_integer)
default_options.redirect_option('log_view_host', 'logview_host')
default_options.redirect_option('log_view_hours', 'logview_hours')
default_options.register_option('api_proxy', None)
default_options.register_option('data_proxy', None)
default_options.redirect_option('tunnel_proxy', 'data_proxy')
default_options.register_option('seahawks_url', None)
default_options.register_option('biz_id', None)
default_options.register_option('priority', None, validator=any_validator(is_null, is_integer))
default_options.register_option('get_priority', None)
default_options.register_option('temp_lifecycle', 1, validator=is_integer)
default_options.register_option('lifecycle', None, validator=any_validator(is_null, is_integer))
default_options.register_option('table_read_limit', None, validator=any_validator(is_null, is_integer))
default_options.register_option('completion_size', 10, validator=is_integer)
default_options.register_option('default_task_settings', None, validator=any_validator(is_null, is_dict))
default_options.register_option('resource_chunk_size', 64 << 20, validator=is_integer)
default_options.register_option('upload_resource_in_chunks', True, validator=is_bool)
default_options.register_option('verify_ssl', True)
default_options.register_option('always_enable_schema', False, validator=is_bool)
default_options.register_option('table_auto_flush_time', 150, validator=is_integer)
default_options.register_option('struct_as_dict', False, validator=is_bool)
default_options.register_option('progress_time_interval', 5 * 60, validator=any_validator(is_float, is_integer))
default_options.register_option('progress_percentage_gap', 5, validator=is_integer)
default_options.register_option('enable_v4_sign', False, validator=is_bool)

# c or python mode, use for UT, in other cases, please do not modify the value
default_options.register_option('force_c', False, validator=is_integer)
default_options.register_option('force_py', False, validator=is_integer)

# callbacks for wrappers
default_options.register_option('instance_create_callback', None)
default_options.register_option("tunnel_session_create_callback", None)
default_options.register_option("tunnel_session_create_timeout_callback", None)
default_options.register_option("result_reader_create_callback", None)
default_options.register_option('tunnel_read_timeout_callback', None)
default_options.register_option("skipped_survey_regexes", [])

# network connections
default_options.register_option('chunk_size', DEFAULT_CHUNK_SIZE, validator=is_integer)
default_options.register_option('retry_times', DEFAULT_CONNECT_RETRY_TIMES, validator=is_integer)
default_options.register_option('retry_delay', DEFAULT_RETRY_DELAY, validator=any_validator(is_integer, is_float))
default_options.register_option('connect_timeout', DEFAULT_CONNECT_TIMEOUT, validator=is_integer)
default_options.register_option('read_timeout', DEFAULT_READ_TIMEOUT, validator=is_integer)
default_options.register_option('pool_connections', DEFAULT_POOL_CONNECTIONS, validator=is_integer)
default_options.register_option('pool_maxsize', DEFAULT_POOL_MAXSIZE, validator=is_integer)

# Tunnel
default_options.register_option('tunnel.endpoint', None)
default_options.register_option('tunnel.string_as_binary', False, validator=is_bool)
default_options.register_option('tunnel.use_instance_tunnel', True, validator=is_bool)
default_options.register_option(
    'tunnel.limit_instance_tunnel', None, validator=any_validator(is_null, is_bool)
)
default_options.register_option(
    'tunnel.legacy_fallback_timeout', None, validator=any_validator(is_null, is_integer)
)
default_options.register_option('tunnel.pd_mem_cache_size', 1024 * 4, validator=is_integer)
default_options.register_option('tunnel.pd_row_cache_size', 1024 * 16, validator=is_integer)
default_options.register_option('tunnel.read_row_batch_size', 1024, validator=is_integer)
default_options.register_option('tunnel.write_row_batch_size', 1024, validator=is_integer)
default_options.register_option('tunnel.batch_merge_threshold', 128, validator=is_integer)
default_options.register_option('tunnel.overflow_date_as_none', False, validator=is_bool)
default_options.register_option(
    'tunnel.quota_name', None, validator=any_validator(is_null, is_string)
)
default_options.register_option('tunnel.block_buffer_size', 20 * 1024 ** 2, validator=is_integer)
default_options.register_option('tunnel.use_block_writer_by_default', False, validator=is_bool)

default_options.redirect_option('tunnel_endpoint', 'tunnel.endpoint')
default_options.redirect_option('use_instance_tunnel', 'tunnel.use_instance_tunnel')
default_options.redirect_option('limited_instance_tunnel', 'tunnel.limit_instance_tunnel')
default_options.redirect_option('tunnel.limited_instance_tunnel', 'tunnel.limit_instance_tunnel')

# terminal
default_options.register_option('console.max_lines', None)
default_options.register_option('console.max_width', None)
default_options.register_option('console.use_color', False, validator=is_bool)

# SQL
default_options.register_option('sql.settings', None, validator=any_validator(is_null, is_dict))
default_options.register_option('sql.ignore_fields_not_null', False, validator=is_bool)
default_options.register_option('sql.use_odps2_extension', None, validator=any_validator(is_null, is_bool))

# sqlalchemy
default_options.register_option('sqlalchemy.project_as_schema', False, validator=is_bool)

# DataFrame
default_options.register_option('interactive', is_interactive(), validator=is_bool)
default_options.register_option('verbose', False, validator=all_validator(is_bool, verbose_log_validator))
default_options.register_option('verbose_log', None)
default_options.register_option('df.optimize', True, validator=is_bool)
default_options.register_option('df.optimizes.cp', True, validator=is_bool)
default_options.register_option('df.optimizes.pp', True, validator=is_bool)
default_options.register_option('df.optimizes.tunnel', True, validator=is_bool)
default_options.register_option('df.analyze', True, validator=is_bool)
default_options.register_option('df.use_cache', True, validator=is_bool)
default_options.register_option('df.quote', True, validator=is_bool)
default_options.register_option('df.dump_udf', False, validator=is_bool)
default_options.register_option('df.supersede_libraries', True, validator=is_bool)
default_options.register_option('df.libraries', None)
default_options.register_option('df.image', None)
default_options.register_option('df.odps.sort.limit', 10000)
default_options.register_option('df.odps.nan_handler', 'py')  # None for not handled, builtin for built-in ISNAN function
default_options.register_option('df.sqlalchemy.execution_options', None, validator=any_validator(is_null, is_dict))
default_options.register_option('df.seahawks.max_size', 10 * 1024 * 1024 * 1024)  # 10G
default_options.register_option('df.delete_udfs', True, validator=is_bool)
default_options.register_option('df.use_xflow_sample', False, validator=is_bool)
default_options.register_option('df.writer_count_limit', 50, validator=is_integer)

# PyODPS ML
default_options.register_option('ml.xflow_project', 'algo_public', validator=is_string)
default_options.register_option('ml.xflow_settings', None, validator=any_validator(is_null, is_dict))
default_options.register_option('ml.dry_run', False, validator=is_bool)
default_options.register_option('ml.use_model_transfer', False, validator=is_bool)
default_options.register_option('ml.use_old_metrics', True, validator=is_bool)
default_options.register_option('ml.model_volume', 'pyodps_volume', validator=is_string)

# Runner
default_options.redirect_option('runner.dry_run', 'ml.dry_run')

# display
from .console import detect_console_encoding

default_options.register_pandas('display.encoding', detect_console_encoding(), validator=is_string)
default_options.register_pandas('display.max_rows', 60, validator=any_validator(is_null, is_integer))
default_options.register_pandas('display.max_columns', 20, validator=any_validator(is_null, is_integer))
default_options.register_pandas('display.large_repr', 'truncate', validator=is_in(['truncate', 'info']))
default_options.register_pandas('display.notebook_repr_html', True, validator=is_bool)
default_options.register_pandas('display.precision', 6, validator=is_integer)
default_options.register_pandas('display.float_format', None)
default_options.register_pandas('display.chop_threshold', None)
default_options.register_pandas('display.column_space', 12, validator=is_integer)
default_options.register_pandas('display.pprint_nest_depth', 3, validator=is_integer)
default_options.register_pandas('display.max_seq_items', 100, validator=is_integer)
default_options.register_pandas('display.max_colwidth', 50, validator=is_integer)
default_options.register_pandas('display.multi_sparse', True, validator=is_bool)
default_options.register_pandas('display.colheader_justify', 'right', validator=is_string)
default_options.register_pandas('display.unicode.ambiguous_as_wide', False, validator=is_bool)
default_options.register_pandas('display.unicode.east_asian_width', False, validator=is_bool)
default_options.redirect_option('display.height', 'display.max_rows')
default_options.register_pandas('display.width', 80, validator=any_validator(is_null, is_integer))
default_options.register_pandas('display.expand_frame_repr', True)
default_options.register_pandas('display.show_dimensions', 'truncate', validator=is_in([True, False, 'truncate']))

default_options.register_option('display.notebook_widget', True, validator=is_bool)
default_options.redirect_option('display.notebook_repr_widget', 'display.notebook_widget')

# Mars
default_options.register_option('mars.use_common_proxy', True, validator=is_bool)
default_options.register_option('mars.launch_notebook', False, validator=is_bool)
default_options.register_option('mars.to_dataframe_memory_scale', None, validator=any_validator(is_null, is_integer))
default_options.register_option('mars.container_status_timeout', 120, validator=is_integer)


_options_local = threading.local()
if contextvars is not None:
    _options_ctx_var = contextvars.ContextVar("_options_ctx_var")
else:
    _options_ctx_var = None


def reset_global_options():
    global _options_local, _options_ctx_var
    if _options_ctx_var is not None:
        _options_ctx_var = contextvars.ContextVar("_options_ctx_var")
        _options_ctx_var.set(default_options)
    else:
        _options_local = threading.local()
        _options_local.default_options = default_options


reset_global_options()


def get_global_options(copy=False):
    if _options_ctx_var is not None:
        ret = _options_ctx_var.get(None)
    else:
        ret = getattr(_options_local, "default_options", None)

    if ret is None:
        if not copy:
            ret = default_options
        else:
            ret = Config(deepcopy(default_options._config))
        if _options_ctx_var is not None:
            _options_ctx_var.set(ret)
        else:
            _options_local.default_options = ret

    return ret


def set_global_options(opts):
    if _options_ctx_var is not None:
        _options_ctx_var.set(opts)
    else:
        _options_local.default_options = opts


@contextlib.contextmanager
def option_context(config=None):
    global_options = get_global_options(copy=True)

    try:
        config = config or dict()
        local_options = Config(deepcopy(global_options._config))
        local_options.update(config)
        set_global_options(local_options)
        yield local_options
    finally:
        set_global_options(global_options)


class OptionsProxy(object):
    def __dir__(self):
        return dir(get_global_options())

    def __getattribute__(self, attr):
        return getattr(get_global_options(), attr)

    def __setattr__(self, key, value):
        setattr(get_global_options(), key, value)


options = OptionsProxy()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import copy
import logging
import sys
import threading
import time
import types as tps
import uuid  # don't remove
import warnings

from ....compat import six
from ....config import options
from ....errors import ODPSError, NoPermission, ConnectTimeout, ParseError
from ....utils import TEMP_TABLE_PREFIX
from ....models import Table, TableSchema
from ....models.partition import Partition
from ....tempobj import register_temp_table
from ....ui import reload_instance_status, fetch_instance_group
from ...core import DataFrame
from ...expr.core import ExprDAG
from ...expr.dynamic import DynamicMixin
from ...expr.expressions import CollectionExpr, Scalar, Summary, SequenceExpr
from ...utils import is_source_collection, is_constant_scalar
from ...types import DynamicSchema, Unknown
from ..utils import refresh_dynamic, process_persist_kwargs
from ..core import Engine, ExecuteNode
from ..errors import CompileError
from ..frame import ResultFrame
from ..context import context
from . import types
from . import analyzer as ana
from . import rewriter as rwr
from .context import ODPSContext, UDF_CLASS_NAME
from .compiler import OdpsSQLCompiler
from .codegen import gen_udf
from .tunnel import TunnelEngine
from .models import MemCacheReference

logger = logging.getLogger(__name__)


class SQLExecuteNode(ExecuteNode):
    def _sql(self):
        raise NotImplementedError

    def __repr__(self):
        buf = six.StringIO()

        sql = self._sql()

        if sql:
            if isinstance(sql, list):
                sql = '\n'.join(sql)
            buf.write('SQL compiled: \n\n')
            buf.write(sql)
        else:
            buf.write('Use tunnel to download data')

        return buf.getvalue()

    def _repr_html_(self):
        buf = six.StringIO()

        sql = self._sql()

        if sql:
            if isinstance(sql, list):
                sql = '\n'.join(sql)
            buf.write('<h4>SQL compiled</h4>')
            buf.write('<code>%s</code>' % sql)
        else:
            buf.write('<p>Use tunnel to download data</p>')

        return buf.getvalue()


class ODPSSQLEngine(Engine):
    def __init__(self, odps):
        self._odps = odps
        self._ctx_local = threading.local()
        self._instances = []

    @property
    def _ctx(self):
        if not hasattr(self._ctx_local, '_ctx'):
            self._ctx_local._ctx = ODPSContext(self._odps)
        return self._ctx_local._ctx

    def stop(self):
        for inst_id in self._instances:
            try:
                self._odps.stop_instance(inst_id)
            except ODPSError:
                pass
        self._ctx.close()

    def _new_execute_node(self, expr_dag):
        node = SQLExecuteNode(expr_dag)

        def _sql(*_):
            return self._compile_sql(node.expr_dag)

        node._sql = tps.MethodType(_sql, node)
        return node

    @staticmethod
    def _get_task_percent(task_progress):
        if len(task_progress.stages) > 0:
            all_percent = sum((float(stage.terminated_workers) / stage.total_workers)
                              for stage in task_progress.stages if stage.total_workers > 0)
            return all_percent / len(task_progress.stages)
        else:
            return 0

    def _reload_ui(self, group, instance, ui):
        if group:
            reload_instance_status(self._odps, group, instance.id)
            ui.update_group()
            return fetch_instance_group(group).instances.get(instance.id)

    def _run(self, sql, ui, progress_proportion=1, hints=None, priority=None,
             running_cluster=None, group=None, libraries=None, schema=None, image=None):
        libraries = self._ctx.prepare_resources(self._get_libraries(libraries))
        self._ctx.create_udfs(libraries=libraries)

        hints = hints or dict()
        if self._ctx.get_udf_count() > 0 and sys.version_info[:2] >= (3, 6):
            hints['odps.sql.jobconf.odps2'] = True
            if sys.version_info[:2] >= (3, 11):
                hints['odps.sql.python.version'] = 'cp311'
            else:
                hints['odps.sql.python.version'] = 'cp37'
        image = image or options.df.image
        if image:
            hints['odps.session.image'] = image

        instance = self._odps.run_sql(sql, hints=hints, priority=priority, name='PyODPSDataFrameTask',
                                      running_cluster=running_cluster, default_schema=schema)

        self._instances.append(instance.id)
        logger.info(
            'Instance ID: %s\n  Log view: %s', instance.id, instance.get_logview_address()
        )
        ui.status('Executing', 'execution details')

        percent = 0
        last_log_progress = 0
        progress_time = start_time = time.time()
        while not instance.is_terminated(retry=True):
            inst_progress = self._reload_ui(group, instance, ui)

            if inst_progress:
                last_percent = percent
                if inst_progress is not None and len(inst_progress.tasks) > 0:
                    percent = sum(self._get_task_percent(task)
                                  for task in six.itervalues(inst_progress.tasks)) \
                              / len(inst_progress.tasks)
                else:
                    percent = 0
                percent = min(1, max(percent, last_percent))
                ui.inc((percent - last_percent) * progress_proportion)

            if logger.getEffectiveLevel() <= logging.INFO:
                try:
                    check_time = time.time()
                    task_progresses = inst_progress.tasks
                    total_progress = sum(
                        stage.finished_percentage
                        for progress in task_progresses.values()
                        for stage in progress.stages
                    )
                    if check_time - start_time >= options.progress_time_interval and (
                        total_progress - last_log_progress >= options.progress_percentage_gap
                        or check_time - progress_time >= options.progress_time_interval
                    ):
                        output_parts = [instance.id] + [
                            progress.get_stage_progress_formatted_string()
                            for progress in task_progresses.values()
                        ]
                        logger.info(" ".join(output_parts))
                        last_log_progress = total_progress
                        progress_time = check_time
                except:
                    # make sure progress display does not affect execution
                    pass

            time.sleep(1)

        instance.wait_for_success()

        self._reload_ui(group, instance, ui)
        if percent < 1:
            ui.inc((1 - percent) * progress_proportion)

        return instance

    def _handle_cases(self, *args, **kwargs):
        tunnel_engine = TunnelEngine(self._odps)
        return tunnel_engine.execute(*args, **kwargs)

    def _gen_table_name(self):
        table_name = '%s%s_%s' % (TEMP_TABLE_PREFIX, int(time.time()),
                                  str(uuid.uuid4()).replace('-', '_'))
        register_temp_table(self._odps, table_name, schema=self._ctx.default_schema)
        return table_name

    def _new_analyzer(self, expr_dag, on_sub=None):
        return ana.Analyzer(expr_dag, on_sub=on_sub)

    def _new_rewriter(self, expr_dag):
        return rwr.Rewriter(expr_dag)

    def _compile_sql(self, expr_dag, prettify=True):
        self._rewrite(expr_dag)

        return self._compile(expr_dag.root, prettify=prettify)

    def _compile(self, expr, prettify=False, libraries=None):
        backend = OdpsSQLCompiler(self._ctx, beautify=prettify)

        libraries = self._ctx.prepare_resources(self._get_libraries(libraries))
        self._ctx.register_udfs(*gen_udf(expr, UDF_CLASS_NAME, libraries=libraries))

        return backend.compile(expr)

    def _mem_cache(self, expr_dag, expr):
        engine = self
        root = expr

        def _sub():
            ref_name = self._ctx.get_mem_cache_ref_name(root)
            sub = CollectionExpr(_source_data=MemCacheReference(root._id, ref_name),
                                 _schema=expr_dag.root.schema, _id=expr_dag.root._id)
            expr_dag.substitute(expr_dag.root, sub)
            return sub

        if self._ctx.is_expr_mem_cached(expr):
            if is_source_collection(expr) and \
                    isinstance(expr._source_data, MemCacheReference):
                return
            _sub()
            return

        class MemCacheCompiler(OdpsSQLCompiler):
            def visit_source_collection(self, expr):
                if isinstance(expr._source_data, MemCacheReference):
                    alias = self._ctx.register_collection(expr)
                    from_clause = '{0} {1}'.format(expr._source_data.ref_name, alias)
                    self.add_from_clause(expr, from_clause)
                    self._ctx.add_expr_compiled(expr, from_clause)
                    engine._ctx.register_mem_cache_dep(root, expr)
                else:
                    super(MemCacheCompiler, self).visit_source_collection(expr)

        compiler = MemCacheCompiler(self._ctx, indent_size=0)
        sql = compiler.compile(expr_dag.root).replace('\n', '')
        self._ctx.register_mem_cache_sql(root, sql)
        sub = _sub()

        return sub, None

    def _cache(self, expr_dag, dag, expr, **kwargs):
        if isinstance(expr, CollectionExpr) and expr._mem_cache:
            return self._mem_cache(expr_dag, expr)

        # prevent the kwargs come from `persist`
        process_persist_kwargs(kwargs)

        if is_source_collection(expr_dag.root) or \
                is_constant_scalar(expr_dag.root):
            return

        execute_dag = ExprDAG(expr_dag.root, dag=expr_dag)

        if isinstance(expr, CollectionExpr):
            table_name = self._gen_table_name()
            table = self._odps.get_table(table_name, schema=self._ctx.default_schema)
            root = expr_dag.root
            sub = CollectionExpr(_source_data=table, _schema=expr.schema)
            sub.add_deps(root)
            expr_dag.substitute(root, sub)

            kw = dict(kwargs)
            kw['lifecycle'] = options.temp_lifecycle

            execute_node = self._persist(table_name, execute_dag, dag, expr, **kw)

            def callback(_):
                if isinstance(expr, DynamicMixin):
                    sub._schema = types.odps_schema_to_df_schema(table.table_schema)
                    refresh_dynamic(sub, expr_dag)

            execute_node.callback = callback
        else:
            assert isinstance(expr, Scalar)  # sequence is not cache-able

            class ValueHolder(object): pass
            sub = Scalar(_value_type=expr.dtype)
            sub._value = ValueHolder()
            root = expr_dag.root
            sub.add_deps(root)
            expr_dag.substitute(root, sub)

            execute_node = self._execute(execute_dag, dag, expr, **kwargs)

            def callback(res):
                sub._value = res
            execute_node.callback = callback

        return sub, execute_node

    @classmethod
    def _join_sql(cls, sql):
        if isinstance(sql, list):
            return '\n'.join(sql)
        return sql

    def _do_execute(self, expr_dag, expr, ui=None, progress_proportion=1,
                    lifecycle=None, head=None, tail=None, hints=None,
                    priority=None, running_cluster=None, schema=None, **kw):
        lifecycle = lifecycle or options.temp_lifecycle
        group = kw.get('group')
        libraries = kw.pop('libraries', None)
        image = kw.pop('image', None)
        use_tunnel = kw.get('use_tunnel', True)

        self._ctx.default_schema = schema or self._ctx.default_schema
        if self._odps.is_schema_namespace_enabled(hints):
            self._ctx.default_schema = self._ctx.default_schema or "default"

        expr_dag = self._convert_table(expr_dag)
        self._rewrite(expr_dag)

        src_expr = expr
        expr = expr_dag.root

        if isinstance(expr, Scalar) and expr.value is not None:
            ui.inc(progress_proportion)
            return expr.value

        no_permission = False
        if options.df.optimizes.tunnel:
            force_tunnel = kw.get('_force_tunnel', False)
            try:
                result = self._handle_cases(expr, ui, progress_proportion=progress_proportion,
                                            head=head, tail=tail)
            except KeyboardInterrupt:
                ui.status('Halt by interruption')
                sys.exit(1)
            except (NoPermission, ConnectTimeout) as ex:
                result = None
                no_permission = True
                if head:
                    expr = expr[:head]
                warnings.warn('Failed to download data by table tunnel, 10000 records will be limited.\n' +
                              'Cause: ' + str(ex))
            if force_tunnel or result is not None:
                return result

        try:
            sql = self._compile(expr, libraries=libraries)
            if types.get_local_use_odps2_types(self._odps.get_project()):
                hints = copy.copy(hints or {})
                hints["odps.sql.type.system.odps2"] = "true"

            cache_data = None
            if not no_permission and isinstance(expr, CollectionExpr) and not isinstance(expr, Summary):
                # When tunnel cannot handle, we will try to create a table
                tmp_table_name = '%s%s' % (TEMP_TABLE_PREFIX, str(uuid.uuid4()).replace('-', '_'))
                register_temp_table(self._odps, tmp_table_name, schema=self._ctx.default_schema)
                cache_data = self._odps.get_table(tmp_table_name, schema=self._ctx.default_schema)

                lifecycle_str = 'LIFECYCLE {0} '.format(lifecycle) if lifecycle is not None else ''
                format_sql = lambda s: 'CREATE TABLE {0} {1}AS \n{2}'.format(tmp_table_name, lifecycle_str, s)
                if isinstance(sql, list):
                    sql[-1] = format_sql(sql[-1])
                else:
                    sql = format_sql(sql)

            sql = self._join_sql(sql)

            logger.info('Sql compiled:\n' + sql)

            try:
                instance = self._run(sql, ui, progress_proportion=progress_proportion * 0.9, hints=hints,
                                     priority=priority, running_cluster=running_cluster, group=group,
                                     libraries=libraries, image=image, schema=self._ctx.default_schema)
            finally:
                self._ctx.close()  # clear udfs and resources generated

            res = self._fetch(expr, src_expr, instance, ui,
                              cache_data=cache_data, head=head, tail=tail,
                              use_tunnel=use_tunnel, group=group,
                              progress_proportion=progress_proportion * 0.1,
                              finish=kw.get('finish', True))
        finally:
            types.set_local_use_odps2_types(None)

        if kw.get('ret_instance', False) is True:
            return instance, res
        return res

    def _fetch(self, expr, src_expr, instance, ui, progress_proportion=1,
               cache_data=None, head=None, tail=None, use_tunnel=True,
               group=None, finish=True):
        if isinstance(expr, (CollectionExpr, Summary)):
            df_schema = expr._schema
            schema = types.df_schema_to_odps_schema(expr._schema, ignorecase=True)
        elif isinstance(expr, SequenceExpr):
            df_schema = TableSchema.from_lists([expr.name], [expr._data_type])
            schema = types.df_schema_to_odps_schema(df_schema, ignorecase=True)
        else:
            df_schema = None
            schema = None

        if cache_data is not None:
            if group and finish:
                ui.remove_keys(group)
            if use_tunnel:
                try:
                    if finish:
                        ui.status('Start to use tunnel to download results...')
                    with cache_data.open_reader(reopen=True) as reader:
                        if head:
                            reader = reader[:head]
                        elif tail:
                            start = max(reader.count - tail, 0)
                            reader = reader[start: ]
                        try:
                            return ResultFrame([r.values for r in reader], schema=df_schema)
                        finally:
                            context.cache(src_expr, cache_data)
                            # reset schema
                            if isinstance(src_expr, CollectionExpr) and \
                                    (isinstance(src_expr._schema, DynamicSchema) or
                                     any(isinstance(col.type, Unknown) for col in src_expr._schema.columns)):
                                src_expr._schema = df_schema
                            ui.inc(progress_proportion)
                except ODPSError as ex:
                    # some project has closed the tunnel download
                    # we just ignore the error
                    warnings.warn('Failed to download data by table tunnel, 10000 records will be limited.\n' +
                                  'Cause: ' + str(ex))
                    pass

            if tail:
                raise NotImplementedError

            try:
                if finish:
                    ui.status('Start to use head to download results...')
                return ResultFrame(cache_data.head(head or 10000), schema=df_schema)
            finally:
                context.cache(src_expr, cache_data)
                ui.inc(progress_proportion)

        with instance.open_reader(schema=schema, use_tunnel=False) as reader:
            ui.status('Start to read instance results...')
            if not isinstance(src_expr, Scalar):
                if head:
                    reader = reader[:head]
                elif tail:
                    start = max(reader.count - tail, 0)
                    reader = reader[start: ]
                try:
                    return ResultFrame([r.values for r in reader], schema=df_schema)
                finally:
                    context.cache(src_expr, cache_data)
                    ui.inc(progress_proportion)
            else:
                ui.inc(progress_proportion)
                odps_type = types.df_type_to_odps_type(src_expr._value_type, project=instance.project)
                res = types.odps_types.validate_value(reader[0][0], odps_type)
                context.cache(src_expr, res)
                return res

    def _do_persist(self, expr_dag, expr, name, partitions=None, partition=None, project=None, ui=None,
                    progress_proportion=1, lifecycle=None, hints=None, priority=None,
                    running_cluster=None, overwrite=True, drop_table=False, create_table=True,
                    drop_partition=False, create_partition=None, cast=False, schema=None, **kw):
        group = kw.get('group')
        libraries = kw.pop('libraries', None)
        image = kw.pop('image', None)

        if isinstance(name, Partition):
            partition = name.partition_spec
            name = name.table
        if isinstance(name, Table):
            table = name
            project = table.project.name
            if table.get_schema():
                schema = table.get_schema().name
            name = table.name

        lifecycle = options.temp_lifecycle if name.startswith(TEMP_TABLE_PREFIX) else lifecycle

        self._ctx.default_schema = schema or self._ctx.default_schema
        if self._odps.is_schema_namespace_enabled(hints):
            self._ctx.default_schema = self._ctx.default_schema or "default"

        expr_dag = self._convert_table(expr_dag)
        self._rewrite(expr_dag)

        src_expr = expr
        expr = expr_dag.root

        should_cache = False

        if drop_table:
            self._odps.delete_table(name, project=project, schema=schema, if_exists=True)

        if project is not None or schema is not None:
            project = project or self._ctx._odps.project
            schema = schema or self._ctx.default_schema
        if project is None:
            table_name = name
        elif schema is None:
            table_name = '%s.`%s`' % (project, name)
        else:
            table_name = '%s.%s.`%s`' % (project, schema, name)

        project_obj = self._odps.get_project(project)
        if partitions is None and partition is None:
            # the non-partitioned table
            if drop_partition:
                raise ValueError('Cannot drop partition for non-partition table')
            if create_partition:
                raise ValueError('Cannot create partition for non-partition table')

            if self._odps.exist_table(name, project=project, schema=schema) or not create_table:
                t = self._odps.get_table(name, project=project, schema=schema)
                if t.table_schema.partitions:
                    raise CompileError('Cannot insert into partition table %s without specifying '
                                       '`partition` or `partitions`.')
                expr = self._reorder(expr, t, cast=cast)
            else:
                # We don't use `CREATE TABLE ... AS` because it will report `table already exists`
                # when service retries.
                if isinstance(expr, CollectionExpr):
                    t_schema = types.df_schema_to_odps_schema(expr.schema, ignorecase=True)
                else:
                    col_name = expr.name
                    tp = types.df_type_to_odps_type(expr.dtype, project=project_obj)
                    t_schema = TableSchema.from_lists([col_name, ], [tp, ])
                self._odps.create_table(name, TableSchema(columns=t_schema.columns),
                                        project=project, schema=schema, lifecycle=lifecycle)

            sql = self._compile(expr, prettify=False, libraries=libraries)
            action_str = 'OVERWRITE' if overwrite else 'INTO'
            format_sql = lambda s: 'INSERT {0} TABLE {1} \n{2}'.format(action_str, table_name, s)
            if isinstance(sql, list):
                sql[-1] = format_sql(sql[-1])
            else:
                sql = format_sql(sql)

            should_cache = True
        elif partition is not None:
            if self._odps.exist_table(name, project=project, schema=schema) or not create_table:
                t = self._odps.get_table(name, project=project, schema=schema)
                partition = self._get_partition(partition, t)

                if drop_partition:
                    t.delete_partition(partition, if_exists=True)
                if create_partition:
                    t.create_partition(partition, if_not_exists=True)
            else:
                partition = self._get_partition(partition)
                column_names = [n for n in expr.schema.names if n not in partition]
                column_types = [
                    types.df_type_to_odps_type(expr.schema[n].type, project=project_obj)
                    for n in column_names
                ]
                partition_names = [n for n in partition.keys]
                partition_types = ['string'] * len(partition_names)
                t = self._odps.create_table(
                    name, TableSchema.from_lists(
                        column_names, column_types, partition_names, partition_types
                    ),
                    project=project, lifecycle=lifecycle, schema=schema)
                if create_partition is None or create_partition is True:
                    t.create_partition(partition)

            expr = self._reorder(expr, t, cast=cast)
            sql = self._compile(expr, prettify=False, libraries=libraries)

            action_str = 'OVERWRITE' if overwrite else 'INTO'
            format_sql = lambda s: 'INSERT {0} TABLE {1} PARTITION({2}) \n{3}'.format(
                action_str, table_name, partition, s
            )
            if isinstance(sql, list):
                sql[-1] = format_sql(sql[-1])
            else:
                sql = format_sql(sql)
        else:
            if isinstance(partitions, tuple):
                partitions = list(partitions)
            if not isinstance(partitions, list):
                partitions = [partitions, ]

            if isinstance(expr, CollectionExpr):
                t_schema = types.df_schema_to_odps_schema(expr.schema, ignorecase=True)
            else:
                col_name = expr.name
                tp = types.df_type_to_odps_type(expr.dtype, project=project_obj)
                t_schema = TableSchema.from_lists([col_name, ], [tp, ])

            for p in partitions:
                if p not in t_schema:
                    raise ValueError(
                        'Partition field(%s) does not exist in DataFrame schema' % p)

            columns = [c for c in t_schema.columns if c.name not in partitions]
            ps = [TableSchema.TablePartition(name=pt, type=t_schema.get_type(pt)) for pt in partitions]
            if self._odps.exist_table(name, project=project, schema=schema) or not create_table:
                t = self._odps.get_table(name, project=project, schema=schema)
            else:
                t = self._odps.create_table(name, TableSchema(columns=columns, partitions=ps),
                                            project=project, lifecycle=lifecycle, schema=schema)
            if drop_partition:
                raise ValueError('Cannot drop partitions when specify `partitions`')
            if create_partition:
                raise ValueError('Cannot create partitions when specify `partitions`')
            expr = expr[[c.name for c in expr.schema if c.name not in partitions] + partitions]

            expr = self._reorder(expr, t, cast=cast, with_partitions=True)
            sql = self._compile(expr, prettify=False, libraries=libraries)

            action_str = 'OVERWRITE' if overwrite else 'INTO'
            format_sql = lambda s: 'INSERT {0} TABLE {1} PARTITION({2}) \n{3}'.format(
                action_str, table_name, ', '.join(partitions), s
            )
            if isinstance(sql, list):
                sql[-1] = format_sql(sql[-1])
            else:
                sql = format_sql(sql)

        sql = self._join_sql(sql)

        logger.info('Sql compiled:\n' + sql)

        try:
            instance = self._run(sql, ui, progress_proportion=progress_proportion, hints=hints,
                                 priority=priority, running_cluster=running_cluster, group=group,
                                 libraries=libraries, image=image, schema=schema)
        except ParseError as ex:
            logger.error("Failed to run DF generated SQL: %s:\n%s", str(ex), sql)
            raise
        finally:
            self._ctx.close()  # clear udfs and resources generated

        t = self._odps.get_table(name, project=project, schema=schema)
        if should_cache and not is_source_collection(src_expr):
            # TODO: support cache partition
            context.cache(src_expr, t)
        if partition:
            filters = []
            df = DataFrame(t)
            for k in partition.keys:
                # actual type of partition and column type may mismatch
                filters.append(df[k] == Scalar(partition[k]).cast(df[k].dtype))
            res = df.filter(*filters)
        else:
            res = DataFrame(t)
        if kw.get('ret_instance', False) is True:
            return instance, res
        return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..core import Iterable
from ... import serializers, errors
from ...compat import six
from .offlinemodel import OfflineModel


class OfflineModels(Iterable):

    marker = serializers.XMLNodeField('Marker')
    max_items = serializers.XMLNodeField('MaxItems', parse_callback=int)
    offline_models = serializers.XMLNodesReferencesField(OfflineModel, 'OfflineModel')

    @property
    def project(self):
        return self.parent

    def _get(self, item):
        return OfflineModel(client=self._client, parent=self, name=item)

    def __contains__(self, item):
        if isinstance(item, six.string_types):
            offline_model = self._get(item)
        elif isinstance(item, OfflineModel):
            offline_model = item
        else:
            return False

        try:
            offline_model.reload()
            return True
        except errors.NoSuchObject:
            return False

    def __iter__(self):
        return self.iterate()

    def iterate(self, name=None, owner=None):
        """

        :param name: the prefix of offline model name
        :param owner:
        :return:
        """
        params = {'expectmarker': 'true'}
        if name is not None:
            params['name'] = name
        if owner is not None:
            params['owner'] = owner

        def _it():
            last_marker = params.get('marker')
            if 'marker' in params and \
                (last_marker is None or len(last_marker) == 0):
                return

            url = self.resource()
            resp = self._client.get(url, params=params)

            t = OfflineModels.parse(self._client, resp, obj=self)
            params['marker'] = t.marker

            return t.offline_models

        while True:
            offline_models = _it()
            if offline_models is None:
                break
            for offline_model in offline_models:
                yield offline_model

    def delete(self, name):
        if not isinstance(name, OfflineModel):
            offline_model = OfflineModel(name=name, parent=self)
        else:
            offline_model = name
            name = name.name

        del self[name]  # delete from cache

        url = offline_model.resource()

        self._client.delete(url)


# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fnmatch

try:
    import oss2
except ImportError:
    oss2 = None

from ....compat import urlparse


class OSSRandomReader(object):
    mode = "rb"

    def __init__(self, bucket, key):
        self._bucket = bucket
        self._key = key
        self._pos = 0
        self._file_len = None
        self._reader = None
        self._reload_object()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _reload_object(self):
        if self._file_len is None:
            self._file_len = self._bucket.get_object_meta(self._key).content_length
        self._reader = self._bucket.get_object(self._key, (self._pos, None))

    def tell(self):
        return self._pos

    def read(self, amt=None):
        buf = self._reader.read(amt)
        self._pos += len(buf)
        return buf

    def seek(self, offset, whence=0):
        old_pos = self._pos
        if whence == 0:
            self._pos = offset
        elif whence == 1:
            self._pos = min(self._file_len, self._pos + offset)
        else:
            self._pos = max(0, self._file_len - offset)
        if self._pos != old_pos:
            self._reload_object()

    def close(self):
        self._bucket = None
        self._reader = None


class OSSWriter(object):
    mode = "wb"

    def __init__(self, bucket, key):
        self._bucket = bucket
        self._key = key

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def write(self, content):
        self._bucket.put_object(self._key, content)

    def close(self):
        self._bucket = None


class OSSFileSystem(object):
    sep = "/"

    def __init__(
        self, endpoint=None, access_id=None, secret_access_key=None, bucket_name=None
    ):
        self._bucket = oss2.Bucket(
            oss2.Auth(access_id, secret_access_key), endpoint, bucket_name
        )

    def _normalize_path(self, path):
        parse_result = urlparse(path)
        path = parse_result.path
        if parse_result.netloc:
            path = parse_result.netloc + parse_result.path
        return path

    def glob(self, path):
        path = self._normalize_path(path)

        prefix_pos = 0
        while prefix_pos < len(path):
            if str.isalnum(path[prefix_pos]) or path[prefix_pos] in "_/":
                prefix_pos += 1
            else:
                break
        prefix_path = path[0:prefix_pos]

        for obj_info in oss2.ObjectIterator(self._bucket, prefix_path):
            if fnmatch.fnmatch(obj_info.key, path):
                yield "oss://" + obj_info.key

    def open(self, path, mode="rb"):
        path = self._normalize_path(path)
        if mode == "rb":
            return OSSRandomReader(self._bucket, path)
        elif mode == "wb":
            return OSSWriter(self._bucket, path)
        else:
            raise NotImplementedError("Does not support mode except `rb` and `wb`")


from . import core

if oss2:
    core._file_systems["oss"] = OSSFileSystem
